//////////////////////////////////////////////////////////////////////////////
//
// Mesh.h, 12/08/01, Ilja Friedel
//
//
//////////////////////////////////////////////////////////////////////////////

#ifndef MESH_USED
#define MESH_USED

//////////////////////////////////////////////////////////////////////////////

#include <assert.h>
#include <stdio.h>
#include <cmath>
#include <stdlib.h>
#include <iostream>
using namespace std;

#include "../Patch/Patch.h"
#include "../Patch/BicubicPatch.h"
#include "../Patch/ExtraordinaryPatch.h"
#include "../Patch/EigenPatch.h"
#include "../Patch/FirstOrderDataDependentPatch.h"

#include "../Minimizer/DynamicCompCol_Mat_double.h"

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

//#define ITMAX 100
//#define ZEPS  1.0e-15
//#define CGOLD 0.3819660
//#define MOV3(a,b,c,d,e,f) (a)=(d);(b)=(e);(c)=(f);
//#define SIGN(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))
//#define SHFT(a,b,c,d) (a)=(b);(b)=(c);(c)=(d);

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

class MeshPatch;
class Mesh;

class MeshPoint
{
 public:

  int         ID;
  int         constraint;
  //int         valence;
  int         patches_in_ring;
  MeshPatch **ring_table;

  int    load(FILE *f, double *P,int);
  int    save(FILE *f, double *P);
  int    export_IV(char *fname, double *P);

  double X(double *P)
    {
      return(P[3*ID]);
    }

  double Y(double *P)
    {
      return(P[3*ID+1]);
    }

  double Z(double *P)
    {
      return(P[3*ID+2]);
    }

  void set(double *P, double px, double py, double pz)
    {
      P[3*ID]  =px;
      P[3*ID+1]=py;
      P[3*ID+2]=pz;
    }

  void print(double *P)
    {
      cout << "ID="<<ID<<" ["<<X(P)<<", "<<Y(P)<<", "<<Z(P)
	   <<"]"//  valence="<<valence
	   << endl;
      fflush(0);
    }

  MeshPoint(FILE *f, double *P,int);
  ~MeshPoint();
};

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

class MeshPatch
{
 public:

  Patch *patch;
  Mesh  *M;

  int  all_points_constrained;
  int  valence;
  int *points;
  int  boundary_type;
  MeshPatch *next;

  double KmEnergy(double *A);
  void   grad_KmEnergy(double *A, double *grad);
  void   Hess_KmEnergy(double *A, DynamicCompCol_Mat_double *H); 

  double KpEnergy(double *A);
  void   grad_KpEnergy(double *A, double *grad);
  void   Hess_KpEnergy(double *A, DynamicCompCol_Mat_double *H); 
  void   Hess_KpEnergy_mult_vector(double *X, double *A, double *result);

  FirstOrderDataDependentPatch* create_Patch(double *MP);
  void update_Patch(double *MP);

  int load(FILE *f);
  int save(FILE *f);
  int export_IV(char *fname, double *P);

  MeshPatch(Mesh *m, FILE *f);
  MeshPatch(int valence, int *P);
  ~MeshPatch();
};

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

class Mesh
{
 public:
  char   name[1024];
  int    numpoints;
  int    numpatches;

  MeshPoint **MPoints;
  double *P;

  double export_iv_sphere_size;
  int    export_iv_control_grid;
  int    export_iv_M;
  int    export_iv_N;
  int    export_iv_coloring_scheme;
  int    export_iv_reflection_map;
  double export_iv_coloring_range;

//////////////////////////////////////////////////////////////////////////////
// functions that compute energies, derivatives or minima

  double                     KpEnergy(double *A);
  void                       grad_KpEnergy(double *A, double *grad);
  DynamicCompCol_Mat_double *Hess_KpEnergy(double *A); 
  void                       Hess_KpEnergy_mult_vector(double *X, double *A, double *result);

//////////////////////////////////////////////////////////////////////////////

  double                     KmEnergy(double *A);
  void                       grad_KmEnergy(double *A, double *grad);
  DynamicCompCol_Mat_double *Hess_KmEnergy(double *A); 

//////////////////////////////////////////////////////////////////////////////
// structural functions on the mesh
  MeshPatch * firstpatch;
  void set_connectivity_information(void);

  int load(char *fname);
  int save(char *fname);
  int export_IV(char *fname) {return export_IV(fname,P);}
  int export_IV(char *fname, double *P);
  int export_IV(char *fname, double *P, double *grad);

  Mesh(char *fname);
  ~Mesh();
};

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

#endif
