//////////////////////////////////////////////////////////////////////////////
//
// BilinearPatch.h, 07/24/01, Ilja Friedel
//
//
//////////////////////////////////////////////////////////////////////////////

#ifndef BILINEARPATCH_USED
#define BILINEARPATCH_USED

//////////////////////////////////////////////////////////////////////////////

#include <assert.h>
#include <stdio.h>
#include <cmath>
#include <stdlib.h>
#include <iostream>
using namespace std;
#include "Patch.h"
#include "ExtraordinaryPatch.h"

//////////////////////////////////////////////////////////////////////////////

class BilinearPatch:public Patch
{
 public:

  double Kp_function(const double, const double);
  double Km_function(const double, const double);

  double control_points[3][4];

  double evaluate_S(double u, double v, int s)
    {
      assert(u>=0.0 && u<=1.0);
      assert(v>=0.0 && v<=1.0);
      assert(s>=0 && s<=2);
      return((control_points[s][0]*u+control_points[s][1]*(1.0-u))*v+
	     (control_points[s][2]*u+control_points[s][3]*(1.0-u))*(1.0-v));
    }

  double evaluate_D1S(double u, double v, int s)
    {
      assert(u>=0.0 && u<=1.0);
      assert(v>=0.0 && v<=1.0);
      assert(s>=0 && s<=2);
      return((control_points[s][0]-control_points[s][1])*v+
	     (control_points[s][2]-control_points[s][3])*(1.0-v));
    }

  double evaluate_D2S(double u, double v, int s)
    {
      assert(u>=0.0 && u<=1.0);
      assert(v>=0.0 && v<=1.0);
      assert(s>=0 && s<=2);
      return((control_points[s][0]*u+control_points[s][1]*(1.0-u))-
	     (control_points[s][2]*u+control_points[s][3]*(1.0-u)));
    }

  double evaluate_D1D1S(double u, double v, int s)
    {
      assert(u>=0.0 && u<=1.0);
      assert(v>=0.0 && v<=1.0);
      assert(s>=0 && s<=2);
      return(0.0);
    }

  double evaluate_D1D2S(double u, double v, int s)
    {
      assert(u>=0.0 && u<=1.0);
      assert(v>=0.0 && v<=1.0);
      assert(s>=0 && s<=2);
      return((control_points[s][0]-control_points[s][1])-
	     (control_points[s][2]-control_points[s][3]));
    }

  double evaluate_D2D1S(double u, double v, int s)
    {
      assert(u>=0.0 && u<=1.0);
      assert(v>=0.0 && v<=1.0);
      assert(s>=0 && s<=2);
      return((control_points[s][0]-control_points[s][1])-
	     (control_points[s][2]-control_points[s][3]));
    }

  double evaluate_D2D2S(double u, double v, int s)
    {
      assert(u>=0.0 && u<=1.0);
      assert(v>=0.0 && v<=1.0);
      assert(s>=0 && s<=2);
      return(0.0);
    }
  
//////////////////////////////////////////////////////////////////////////////

  void   print(void)
    {
      int i;

      cout << "Bilinear Patch:" << endl;
      for(i=0;i<4;i++)
	{
	  cout << "("
	       << control_points[0][i] << ", "  
	       << control_points[1][i] << ", "  
	       << control_points[2][i] << ")" << endl; 
	}
    }

  void   Kp(Patch *S, double *K);
  void   Km(Patch *S, double *K);

//////////////////////////////////////////////////////////////////////////////

  virtual ~BilinearPatch();
  BilinearPatch(ExtraordinaryPatch *EP);
};

//////////////////////////////////////////////////////////////////////////////

#endif
