//////////////////////////////////////////////////////////////////////////////
//
// ExtraordinaryPatch.h, 07/24/01, Ilja Friedel
//
//////////////////////////////////////////////////////////////////////////////

#include "ExtraordinaryPatch.h"

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

ExtraordinaryPatch::~ExtraordinaryPatch()
{
  if(control_points[0]) delete[] control_points[0];
  if(control_points[1]) delete[] control_points[1];
  if(control_points[2]) delete[] control_points[2];

  control_points[0]=NULL;
  control_points[1]=NULL;
  control_points[2]=NULL;

  if(ES) delete ES;
  ES=NULL;
}

//////////////////////////////////////////////////////////////////////////////

ExtraordinaryPatch::ExtraordinaryPatch(double **data, int N)
{
  int i,j;

  Valence=N;
  //cout << "Valence="<<Valence<<endl;

  if(Valence<3 || Valence>MAX_VALENCE)
    {
      cerr << "Error: Valence="<<Valence<<" out of range!"<<endl;
      exit(1);
    }

  M=2*Valence+8;

  control_points[0]=new double[M];
  control_points[1]=new double[M];
  control_points[2]=new double[M];

  assert(control_points[0]);
  assert(control_points[1]);
  assert(control_points[2]);

  for(i=0;i<3;i++)
    for(j=0;j<M;j++) 
      control_points[i][j]=data[i][j];

  ES=new EigenStructure(this);
}

//////////////////////////////////////////////////////////////////////////////

void ExtraordinaryPatch::export_IV(const char *fn, double r, double g, double b)
{
  int N,M;
  int i,j;
  FILE *f;

  N=SIZE_N;
  M=SIZE_M;

  f=fopen(fn,"a");
  assert(f);
  
  fprintf(f,"#Inventor V2.0 ascii\n\n");
  fprintf(f,"Separator {\n");
  fprintf(f,"	LightModel\n");
  fprintf(f,"	{\n");
  fprintf(f,"	   model BASE_COLOR\n");
  fprintf(f,"	}\n");
  fprintf(f,"	ShapeHints {\n");
  fprintf(f,"	shapeType UNKNOWN_SHAPE_TYPE\n");
  fprintf(f,"	vertexOrdering	COUNTERCLOCKWISE\n");
  fprintf(f,"	}\n");
  fprintf(f,"    PackedColor {\n");
  fprintf(f,"	rgba	[ \n");

  for(i=0;i<=N;i++)
    {
      for(j=0;j<=M;j++)
	{
	  unsigned int  color;
	  unsigned char red, green, blue;
	  
	  if(b<0.0) b=0.0;
	  if(g<0.0) g=0.0;
	  if(r<0.0) r=0.0;

	  if(b>1.0) b=1.0;
	  if(g>1.0) g=1.0;
	  if(r>1.0) r=1.0;

	  blue  = (int)(b*255.0);	  
	  green = (int)(g*255.0);
	  red   = (int)(r*255.0);

	  color=256*(256*(unsigned int)blue + (unsigned int)green) + red;
	      
	  if(i==N && j==M)
	    {
	      fprintf(f,"         0xff%.6x",color);
	    }
	  else
	    {
	      fprintf(f,"         0xff%.6x,",color);
	    }
	}
      fprintf(f,"\n");
    }

  fprintf(f,"]\n");
  fprintf(f,"    }\n");
  fprintf(f,"    MaterialBinding {\n");
  fprintf(f,"	value	PER_VERTEX_INDEXED\n");
  fprintf(f,"    }\n");
  fprintf(f,"    Coordinate3 {\n");
  fprintf(f,"	point	[\n");

  for(i=0;i<=N;i++)
    {
      for(j=0;j<=M;j++)
	{
	  if(i==N && j==M)
	    {
	      fprintf(f,"         %e %e %e",
		      evaluate_S(i/(double)N,j/(double)M, 0),
		      evaluate_S(i/(double)N,j/(double)M, 1),
		      evaluate_S(i/(double)N,j/(double)M, 2));
	    }
	  else
	    {
	      fprintf(f,"         %e %e %e,",
		      evaluate_S(i/(double)N,j/(double)M, 0),
		      evaluate_S(i/(double)N,j/(double)M, 1),
		      evaluate_S(i/(double)N,j/(double)M, 2));
	    }
	  fprintf(f,"\n");
	}
    }

  fprintf(f,"    ]}\n");
  fprintf(f,"    QuadMesh {\n");
  fprintf(f,"	verticesPerColumn	%d\n",N+1);
  fprintf(f,"	verticesPerRow	%d\n",M+1);
  fprintf(f,"    }\n");
  fprintf(f,"}\n");

  fprintf(f,"   Separator {\n");
  fprintf(f,"           Material {\n");
  fprintf(f,"                ambientColor   0.4 0.8 0.4 diffuseColor 0.4 0.8 0.4   \n");
  fprintf(f,"           }\n");
  fprintf(f,"        Coordinate3 {\n");
  fprintf(f,"            point [\n");

  for(i=0;i<4;i++)
    {
      fprintf(f,"%e %e %e",control_points[0][i], control_points[1][i], control_points[2][i]);
      if(i<3) fprintf(f,", ");
      else fprintf(f,"\n");
    }

  fprintf(f,"                  ]\n");
  fprintf(f,"        }\n");
  fprintf(f,"        IndexedLineSet {\n");
  fprintf(f,"            coordIndex [ 0, 1, 2, 3, -1\n");
  fprintf(f,"                       ]\n");
  fprintf(f,"        }");
  fprintf(f,"    }\n");


  fclose(f);
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

double ExtraordinaryPatch::evaluate_S(double u, double v, int s)
{
  int i;
  double tmp;

  assert(s>=0   && s< 3 );
  assert(u>=0.0 && u<=1.0);
  assert(v>=0.0 && v<=1.0);

  tmp=0.0;
  for(i=0;i<M;i++)
    {
      tmp+=ES->evaluate_phi(u,v,i) * ES->eigen_controlpoints[s][i];
    }

 return(tmp);
}

//////////////////////////////////////////////////////////////////////////////

double ExtraordinaryPatch::evaluate_D1S(double u, double v, int s)
{
  int i;
  double tmp;

  assert(s>=0   && s< 3 );
  assert(u>=0.0 && u<=1.0);
  assert(v>=0.0 && v<=1.0);

  tmp=0.0;
  for(i=0;i<M;i++)
    {
      tmp+=ES->evaluate_D1phi(u,v,i) * ES->eigen_controlpoints[s][i];
    }

 return(tmp);
}

//////////////////////////////////////////////////////////////////////////////

double ExtraordinaryPatch::evaluate_D2S(double u, double v, int s)
{
  int i;
  double tmp;

  assert(s>=0   && s< 3 );
  assert(u>=0.0 && u<=1.0);
  assert(v>=0.0 && v<=1.0);

  tmp=0.0;
  for(i=0;i<M;i++)
    {
      tmp+=ES->evaluate_D2phi(u,v,i) * ES->eigen_controlpoints[s][i];
    }

 return(tmp);
}

//////////////////////////////////////////////////////////////////////////////

double ExtraordinaryPatch::evaluate_D1D1S(double u, double v, int s)
{
  int i;
  double tmp;

  assert(s>=0   && s< 3 );
  assert(u>=0.0 && u<=1.0);
  assert(v>=0.0 && v<=1.0);

  tmp=0.0;
  for(i=0;i<M;i++)
    {
      tmp+=ES->evaluate_D1D1phi(u,v,i) * ES->eigen_controlpoints[s][i];
    }

 return(tmp);
}

//////////////////////////////////////////////////////////////////////////////

double ExtraordinaryPatch::evaluate_D1D2S(double u, double v, int s)
{
  int i;
  double tmp;

  assert(s>=0   && s< 3 );
  assert(u>=0.0 && u<=1.0);
  assert(v>=0.0 && v<=1.0);

  tmp=0.0;
  for(i=0;i<M;i++)
    {
      tmp+=ES->evaluate_D1D2phi(u,v,i) * ES->eigen_controlpoints[s][i];
    }

 return(tmp);
}

//////////////////////////////////////////////////////////////////////////////

double ExtraordinaryPatch::evaluate_D2D1S(double u, double v, int s)
{
  int i;
  double tmp;

  assert(s>=0   && s< 3 );
  assert(u>=0.0 && u<=1.0);
  assert(v>=0.0 && v<=1.0);

  tmp=0.0;
  for(i=0;i<M;i++)
    {
      tmp+=ES->evaluate_D2D1phi(u,v,i) * ES->eigen_controlpoints[s][i];
    }

 return(tmp);
}

//////////////////////////////////////////////////////////////////////////////

double ExtraordinaryPatch::evaluate_D2D2S(double u, double v, int s)
{
  int i;
  double tmp;

  assert(s>=0   && s< 3 );
  assert(u>=0.0 && u<=1.0);
  assert(v>=0.0 && v<=1.0);

  tmp=0.0;
  for(i=0;i<M;i++)
    {
      tmp+=ES->evaluate_D2D2phi(u,v,i) * ES->eigen_controlpoints[s][i];
    }

 return(tmp);
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

double ExtraordinaryPatch::Kp_function(const double u, const double v)
{
  double H00, H01, H10, H11;
  double tmp;

  assert(u>=0.0 && u<=1.0);
  assert(v>=0.0 && v<=1.0);

  ReferenceS->set_Hess_S_of_h(u,v,
			      ES->evaluate_D1phi(  u,v,IL_i),
			      ES->evaluate_D2phi(  u,v,IL_i),
			      ES->evaluate_D1D1phi(u,v,IL_i),
			      ES->evaluate_D1D2phi(u,v,IL_i),
			      ES->evaluate_D2D2phi(u,v,IL_i));

  H00 = ReferenceS->Hess_S_of_h[0][0];
  H01 = ReferenceS->Hess_S_of_h[0][1];
  H10 = ReferenceS->Hess_S_of_h[1][0];
  H11 = ReferenceS->Hess_S_of_h[1][1];
  
  ReferenceS->set_Hess_S_of_h(u,v,
			      ES->evaluate_D1phi(  u,v,IL_j),
			      ES->evaluate_D2phi(  u,v,IL_j),
			      ES->evaluate_D1D1phi(u,v,IL_j),
			      ES->evaluate_D1D2phi(u,v,IL_j),
			      ES->evaluate_D2D2phi(u,v,IL_j));
  
  // Greiner:
  // trace(Hess_S(phi(u,v,IL_i))*Hess_S(phi(u,v,IL_j)));
  //tmp = H00 * ReferenceS->Hess_S_of_h[0][0] +
  //      H01 * ReferenceS->Hess_S_of_h[1][0] +  
  //      H10 * ReferenceS->Hess_S_of_h[0][1] +  
  //      H11 * ReferenceS->Hess_S_of_h[1][1];


  // Ilja:
  // trace(Hess_S(phi(u,v,IL_i))*transpose(Hess_S(phi(u,v,IL_j))));
  tmp = H00 * ReferenceS->Hess_S_of_h[0][0] +
        H01 * ReferenceS->Hess_S_of_h[0][1] +  // ******** !!!!!!! *******
        H10 * ReferenceS->Hess_S_of_h[1][0] +  // ********  switched *****
        H11 * ReferenceS->Hess_S_of_h[1][1];

  return(tmp*ReferenceS->Domega);
}

//////////////////////////////////////////////////////////////////////////////

double ExtraordinaryPatch::Km_function(const double u, const double v)
{
  double tmp;
  double G0, G1, G2;

  assert(u>=0.0 && u<=1.0);
  assert(v>=0.0 && v<=1.0);

  ReferenceS->set_Grad_S_of_h(u,v,
			      ES->evaluate_D1phi(u,v,IL_i),
			      ES->evaluate_D2phi(u,v,IL_i));

  G0 = ReferenceS->Grad_S_of_h[0];
  G1 = ReferenceS->Grad_S_of_h[1];
  G2 = ReferenceS->Grad_S_of_h[2];

  ReferenceS->set_Grad_S_of_h(u,v,
			      ES->evaluate_D1phi(u,v,IL_j),
			      ES->evaluate_D2phi(u,v,IL_j));
		     
  // grad_S(phi(u,v,IL_i))*grad_S(phi(u,v,IL_j))

  tmp = G0 * ReferenceS->Grad_S_of_h[0] +
        G1 * ReferenceS->Grad_S_of_h[1] +
        G2 * ReferenceS->Grad_S_of_h[2];

  ReferenceS->set_IIS(); // get Domega

  return(tmp*ReferenceS->Domega);
}

//////////////////////////////////////////////////////////////////////////////

void ExtraordinaryPatch::add_sample_to_Kp(const double u, const double v, const double weight, double *K)
{
  double g11,g12,g22;
  double gg11,gg12,gg22;
  double Gamma1_11, Gamma1_12, Gamma1_22;
  double Gamma2_11, Gamma2_12, Gamma2_22;
  double det_IS;

  double d1h [MAX_VALENCE*2+8];
  double d2h [MAX_VALENCE*2+8];
  double d11h[MAX_VALENCE*2+8];
  double d12h[MAX_VALENCE*2+8];
  double d22h[MAX_VALENCE*2+8];

  double d1S1, d1S2, d1S3;
  double d2S1, d2S2, d2S3;

  double d11S1, d11S2, d11S3;
  double d12S1, d12S2, d12S3;
  double d22S1, d22S2, d22S3;

  double DomegaS;

  int i,j;

  // setting first partial derivatives of surface

  d1S1  = evaluate_D1S(u,v,0);
  d1S2  = evaluate_D1S(u,v,1);
  d1S3  = evaluate_D1S(u,v,2);

  d2S1  = evaluate_D2S(u,v,0);
  d2S2  = evaluate_D2S(u,v,1);
  d2S3  = evaluate_D2S(u,v,2);

  // setting second partial derivatives of surface

  d11S1 = evaluate_D1D1S(u,v,0);
  d11S2 = evaluate_D1D1S(u,v,1);
  d11S3 = evaluate_D1D1S(u,v,2);

  d12S1 = evaluate_D1D2S(u,v,0);
  d12S2 = evaluate_D1D2S(u,v,1);
  d12S3 = evaluate_D1D2S(u,v,2);

  d22S1 = evaluate_D2D2S(u,v,0);
  d22S2 = evaluate_D2D2S(u,v,1);
  d22S3 = evaluate_D2D2S(u,v,2);

  // setting derivatives of eigen basis

  for(i=0;i<M;i++)
    {
      d1h[ i] = ES->evaluate_D1phi(u,v,i);
      d2h[ i] = ES->evaluate_D2phi(u,v,i);
      d11h[i] = ES->evaluate_D1D1phi(u,v,i);
      d12h[i] = ES->evaluate_D1D2phi(u,v,i);
      d22h[i] = ES->evaluate_D2D2phi(u,v,i);
    }

  // setting inverse of first fundamental form IS (times determinant)

  g11 =   d2S1*d2S1 + d2S2*d2S2 + d2S3*d2S3;
  g12 = -(d1S1*d2S1 + d1S2*d2S2 + d1S3*d2S3);
  g22 =   d1S1*d1S1 + d1S2*d1S2 + d1S3*d1S3;

  // determinant of first fundamental form

  det_IS = g11*g22 - g12*g12;

  // setting inverse iIS of first fundamental form IS

  g11 = g11/det_IS;
  g12 = g12/det_IS;
  g22 = g22/det_IS;

  // Christoffel symbols

  Gamma1_11 = g11*(d11S1*d1S1 + d11S2*d1S2 + d11S3*d1S3) + 
              g12*(d11S1*d2S1 + d11S2*d2S2 + d11S3*d2S3);
  Gamma1_12 = g11*(d12S1*d1S1 + d12S2*d1S2 + d12S3*d1S3) + 
              g12*(d12S1*d2S1 + d12S2*d2S2 + d12S3*d2S3);
  Gamma1_22 = g11*(d22S1*d1S1 + d22S2*d1S2 + d22S3*d1S3) + 
              g12*(d22S1*d2S1 + d22S2*d2S2 + d22S3*d2S3);

  Gamma2_11 = g12*(d11S1*d1S1 + d11S2*d1S2 + d11S3*d1S3) + 
              g22*(d11S1*d2S1 + d11S2*d2S2 + d11S3*d2S3);
  Gamma2_12 = g12*(d12S1*d1S1 + d12S2*d1S2 + d12S3*d1S3) + 
              g22*(d12S1*d2S1 + d12S2*d2S2 + d12S3*d2S3);
  Gamma2_22 = g12*(d22S1*d1S1 + d22S2*d1S2 + d22S3*d1S3) + 
              g22*(d22S1*d2S1 + d22S2*d2S2 + d22S3*d2S3);

  // computing iIS &* transpose(iIS)

  gg11 = g11*g11 + g12*g12;
  gg12 = g11*g12 + g12*g22;
  gg22 = g12*g12 + g22*g22;

  // computing DomegaS

  DomegaS=sqrt(det_IS);

  // compare stuff

  set_Hess_S_of_h(u,v,
		  d1h[0],d2h[0],
		  d11h[0],d12h[0],d22h[0]);
  /*
  printf("inv_IS  %e %e %e %e\n",inv_IS[0][0]-g11,
          	                 inv_IS[0][1]-g12,
          	                 inv_IS[1][0]-g12,
  	                         inv_IS[1][1]-g22);

  printf("Gamma1  %e %e %e %e\n",Gamma[0][0][0]-Gamma1_11,
          	                 Gamma[0][1][0]-Gamma1_12,
          	                 Gamma[1][0][0]-Gamma1_12,
  	                         Gamma[1][1][0]-Gamma1_22);

  printf("Gamma2  %e %e %e %e\n",Gamma[0][0][1]-Gamma2_11,
          	                 Gamma[0][1][1]-Gamma2_12,
          	                 Gamma[1][0][1]-Gamma2_12,
  	                         Gamma[1][1][1]-Gamma2_22);

  printf("Domega  %e\n",DomegaS-Domega);
  */

    {
      for(i=0;i<M;i++)
	for(j=0;j<M;j++)
	  {
	    double t0;

	    t0 = 
	      ((d11h[i]-d1h[i]*Gamma1_11-d2h[i]*Gamma2_11)*gg11+
	       (d12h[i]-d1h[i]*Gamma1_12-d2h[i]*Gamma2_12)*gg12)*
	       (d11h[j]-d1h[j]*Gamma1_11-d2h[j]*Gamma2_11)+

	      ((d11h[i]-d1h[i]*Gamma1_11-d2h[i]*Gamma2_11)*gg12+
	       (d12h[i]-d1h[i]*Gamma1_12-d2h[i]*Gamma2_12)*gg22)*
	       (d12h[j]-d1h[j]*Gamma1_12-d2h[j]*Gamma2_12)+

	      ((d12h[i]-d1h[i]*Gamma1_12-d2h[i]*Gamma2_12)*gg11+
	       (d22h[i]-d1h[i]*Gamma1_22-d2h[i]*Gamma2_22)*gg12)*
	       (d12h[j]-d1h[j]*Gamma1_12-d2h[j]*Gamma2_12)+

	      ((d12h[i]-d1h[i]*Gamma1_12-d2h[i]*Gamma2_12)*gg12+
	       (d22h[i]-d1h[i]*Gamma1_22-d2h[i]*Gamma2_22)*gg22)*
	       (d22h[j]-d1h[j]*Gamma1_22-d2h[j]*Gamma2_22);

	    K[i*M+j]+=weight*t0*DomegaS;
	    //K[j*16+i]+=weight*t0*DomegaS;
	  }
    }
}

//////////////////////////////////////////////////////////////////////////////

void ExtraordinaryPatch::add_sample_to_Km(double u, double v, double weight, double *K)
{
  cerr << "void ExtraordinaryPatch::add_sample_to_Km(double u, double v, double weight, double *K)"<< endl;
  assert(u&&v&&weight&&K);
  exit(1);
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

void ExtraordinaryPatch::Kp(Patch *S, double *K)
{
  int    i, j, k;
  double t;
  double *tmp;
  tmp=new double[M*M];

  ReferenceS=S;

  if(S)
    {
      for(IL_i=0;IL_i<M;IL_i++)
	{
	  for(IL_j=IL_i;IL_j<M;IL_j++)
	    {
	      double tmp_I;
	      tmp_I=Integrate((F2D)&ExtraordinaryPatch::Kp_function,
			      0.0,1.0, 0.0,1.0);
	      
	      K[M*IL_j+IL_i]=tmp_I;
	      K[M*IL_i+IL_j]=tmp_I;
	      
	      cout<<".";
	      fflush(0);
	    }
	  cout<<endl;
	}
    }
  else
    {
       for(i=0;i<M*M;i++) K[i]=0.0;

       //add_sample_to_Kp(0.5, 0.5, 1.0, K);

       //add_sample_to_Kp(0.0, 0.0, 0.25, K);
       //add_sample_to_Kp(0.0, 1.0, 0.25, K);
       //add_sample_to_Kp(1.0, 0.0, 0.25, K);
       //add_sample_to_Kp(1.0, 1.0, 0.25, K);
       
       add_sample_to_Kp(    0.5,     0.0, 1.0/12.0, K);
       add_sample_to_Kp(    0.5,     0.5, 1.0/12.0, K);
       add_sample_to_Kp(    1.0,     0.0, 1.0/12.0, K);
       add_sample_to_Kp(    1.0,     0.5, 1.0/12.0, K);

       add_sample_to_Kp(    0.5,     0.5, 1.0/12.0, K);
       add_sample_to_Kp(    0.5,     1.0, 1.0/12.0, K);
       add_sample_to_Kp(    1.0,     0.5, 1.0/12.0, K);
       add_sample_to_Kp(    1.0,     1.0, 1.0/12.0, K);

       add_sample_to_Kp(    0.0,     0.5, 1.0/12.0, K);
       add_sample_to_Kp(    0.0,     1.0, 1.0/12.0, K);
       add_sample_to_Kp(    0.5,     0.5, 1.0/12.0, K);
       add_sample_to_Kp(    0.5,     1.0, 1.0/12.0, K);
 
       for(i=3;i<M;i++)
	 for(j=3;j<M;j++)
	   {	     
	     K[i*M+j]/=(1.0-Lambda[Valence][i*M+i]*Lambda[Valence][j*M+j]/
			    Lambda[Valence][1*M+1]);
	   }
   }

  ReferenceS=NULL;

  // K=mult(transpose(iX)*K*iX);
  ///* ********************************************************
  for(i=0;i<M;++i)
    {
      for(k=0;k<M;++k) 
	{
	  t=0.0;
	  for(j=0;j<M;++j )
	    {
	      t+=K[i*M+j]*iX[Valence][j*M+k];
	    }
	  tmp[i*M+k]=t;
	}
    }
  for(i=0;i<M;++i)
    {
      for(k=0;k<M;++k) 
	{
	  t=0.0;
	  for(j=0;j<M;++j )
	    {
	      t+=iX[Valence][j*M+i]*tmp[j*M+k];
	    } //             ^^^^^ j and i transposed
	  K[i*M+k]=t;
	}
    }
  //*/
}

//////////////////////////////////////////////////////////////////////////////
// ********************
// change it according to Kp!!!
// to do
// ********************

void ExtraordinaryPatch::Km(Patch *S, double *K)
{
  int    i, j, k;
  double t;
  double *tmp;
  tmp=new double[M*M];

  ReferenceS=S;

  for(IL_i=0;IL_i<M;IL_i++)
    for(IL_j=IL_i;IL_j<M;IL_j++)
      {
	K[M*IL_i+IL_j]=Integrate((F2D)&ExtraordinaryPatch::Km_function,
				 0.0,1.0, 0.0,1.0);
	K[M*IL_j+IL_i]=K[M*IL_i+IL_j];
      }

  // K=mult(transpose(iX)*K*iX);
  for(i=0;i<M;++i)
    {
      for(k=0;k<M;++k) 
	{
	  t=0.0;
	  for(j=0;j<M;++j )
	    {
	      t+=K[i*M+j]*iX[Valence][j*M+k];
	    }
	  tmp[i*M+k]=t;
	}
    }
  for(i=0;i<M;++i)
    {
      for(k=0;k<M;++k) 
	{
	  t=0.0;
	  for(j=0;j<M;++j )
	    {
	      t+=iX[Valence][j*M+i]*tmp[j*M+k];
	    } //             ^^^^^ j and i transposed
	  K[i*M+k]=t;
	}
    }
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

void ExtraordinaryPatch::print(void)
{
  cout << "ExtraordinaryPatch  Valence=" << Valence <<endl;
  cout << "ExtraordinaryPatch  M=" << M <<endl;
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// compute Kp matrix and from that get the energy approximation

double ExtraordinaryPatch::KpEnergy(void)
{
  double K[(2*MAX_VALENCE+8)*(2*MAX_VALENCE+8)];
  double E;
  int i,j;

  E=0.0;
  
  Kp(NULL,K);
  for(i=0;i<2*Valence+8;i++)
    {
      for(j=i+1;j<2*Valence+8;j++)
	{
	  E=E+K[(2*Valence+8)*i+j]*
	    (control_points[0][i]*control_points[0][j]+
	     control_points[1][i]*control_points[1][j]+
	     control_points[2][i]*control_points[2][j]);
	}
    }
  E=E*2.0;
  for(i=0;i<2*Valence+8;i++)
     {
	  E=E+K[(2*Valence+8)*i+i]*
	    (control_points[0][i]*control_points[0][i]+
	     control_points[1][i]*control_points[1][i]+
	     control_points[2][i]*control_points[2][i]);
     }

  //cout << "Energy="<<E<<endl;
  
  return E;
}

//////////////////////////////////////////////////////////////////////////////
/*
double ExtraordinaryPatch::grad_KpEnergy(int coord, int i)
{
  ExtraordinaryPatch *EP1;
  ExtraordinaryPatch *EP2;
  double diff;

  assert(0<=coord && coord<3);
  assert(0<=i     && i    <2*Valence+8);

  control_points[coord][i]+=EPSILON3;  // ********************
  EP1=new ExtraordinaryPatch(control_points, Valence);

  control_points[coord][i]-=2*EPSILON3;
  EP2=new ExtraordinaryPatch(control_points, Valence);

  control_points[coord][i]+=EPSILON3;

  diff=(EP1->KpEnergy() - EP2->KpEnergy())/(EPSILON3);

  delete(EP1);
  delete(EP2);

  return(diff);
}
*/
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
