//////////////////////////////////////////////////////////////////////////////
//
// DynamicCompCol_Mat_double.cpp
//
//
//////////////////////////////////////////////////////////////////////////////

#include <iostream.h>
#include <stdlib.h>
#include "DynamicCompCol_Mat_double3x3.h"

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

DCCMdElement3x3::DCCMdElement3x3(DCCMdElement3x3 *DE, int i, double *v)
{
  next=DE;
  row=i;
  val[0]=v[0];
  val[1]=v[1];
  val[2]=v[2];
  val[3]=v[3];
  val[4]=v[4];
  val[5]=v[5];
  val[6]=v[6];
  val[7]=v[7];
  val[8]=v[8];
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

DynamicCompCol_Mat_double3x3::~DynamicCompCol_Mat_double3x3()
{
  int i;
  
  for(i=0;i<dim3_;i++)
    {
      DCCMdElement3x3 *next, *old;

      next=col[i];
      while(next)
	{
	  old=next;
	  next=old->next;
	  delete old;
	}
      col[i]=NULL;
    }

  delete[] col;
}

//////////////////////////////////////////////////////////////////////////////

DynamicCompCol_Mat_double3x3::DynamicCompCol_Mat_double3x3(int d3)
{
  int i;
  dim3_        = d3;
  NumNonZeros_ = 0;
  col          = new DCCMdElement3x3*[dim3_];
  for(i=0;i<dim3_;i++) col[i]=NULL;
}


//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
#ifdef UNDEFINED
int* DynamicCompCol_Mat_double::row_ind(void)
{
  int *r;
  int i,k;
  DCCMdElement *tmp;

  r=new int[NumNonZeros_];
  k=0;

  for(i=0;i<dim_;i++)
    {
      tmp=col[i];
      while(tmp)
	{
	  r[k]=tmp->row;
	  k++;
	  tmp=tmp->next;
	}
    }

  return(r);
}

//////////////////////////////////////////////////////////////////////////////

int* DynamicCompCol_Mat_double::col_ptr(void)
{
  int *c;
  int i,k;
  DCCMdElement *tmp;

  c=new int[dim_+1];
  k=0;

  for(i=0;i<dim_;i++)
    {
      c[i]=k;
      tmp=col[i];
      while(tmp)
	{
	  k++;
	  tmp=tmp->next;
	}
    }
  c[dim_]=NumNonZeros_;

  return(c); 
}

//////////////////////////////////////////////////////////////////////////////

double* DynamicCompCol_Mat_double::val(void)
{
  double *v;
  int     i,k;
  DCCMdElement *tmp;

  v=new double[NumNonZeros_];
  k=0;

  for(i=0;i<dim_;i++)
    {
      tmp=col[i];
      while(tmp)
	{
	  v[k]=tmp->val;
	  k++;
	  tmp=tmp->next;
	}
    }

  return(v);
}
#endif
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
#ifdef UNDEFINED
void DynamicCompCol_Mat_double::save(char *fname)
{
  FILE *f;
  int i,j;
  
  f=fopen(fname,"w");
  assert(f);
  fprintf(f,"P6\n");
  fprintf(f,"# Visualized matrix. Ilja Friedel, March 2002.\n");
  fprintf(f,"%d %d\n", dim(), dim());
  fprintf(f,"255\n");
  
  for(i=0;i<dim();i++)
    for(j=0;j<dim();j++)
      {
	unsigned int r,g,b;
	int x,y;

	r=0;
	g=0;
	b=0;
	
	x=(3*i)%dim()+(3*i)/dim();
	y=(3*j)%dim()+(3*j)/dim();
	

	if((*this)(x,y)> 0.0) 
	  {
	    if((*this)(x,y)> 100.0)
	      {
		r= 255;
	      }
	    else
	      {
		r=128;
		g=0;
		b=0;
	      }
	  }
	if((*this)(x,y)< 0.0)
	  {
	    if((*this)(x,y)<-100.0)
	      {
		b= 255;
	      }
	    else
	      {
		b=128;
		r=0;
		g=0;
	      }
	  }
	if(fabs((*this)(x,y))==1.0) 
	  {
	    r=0;
	    g=255;
	    b=0;
	  }
	if(fabs((*this)(x,y))==0.0)
	  {
	    r=255;
	    g=255;
	    b=255;
	  }
	
	if(r>255) r=255;
	if(b>255) b=255;

	//if((*this)(x,y)== 0.0) 
	// {
	//  r=255-r;
	//  g=255-g;
	//  b=255-b;
	//}
	
	fprintf(f,"%c%c%c",r,g,b);
      }

  fclose(f);
}

//////////////////////////////////////////////////////////////////////////////

void DynamicCompCol_Mat_double::clear_from_zeros(void)
{
  int i;
  
  cout << "Clear: before " << NumNonZeros() << endl;
  
  for(i=0;i<dim();i++)
    {
      DCCMdElement *next, *old;
      
      next=col[i];
      while(next)
	{
	  old=next; // ignores the first element in the list
	  next=old->next;
	  
	  while(next && (fabs(next->val)==0.0))
	    {
	      old->next=next->next;
	      delete next;
	      NumNonZeros_--;
	      next=old->next;
	    }
	}
    }
  cout << "Clear: after " << NumNonZeros() << endl;
}
#endif
//////////////////////////////////////////////////////////////////////////////
// compute matrix-vector multiplication, e.g. A*x=result

void DynamicCompCol_Mat_double3x3::mult(double *x, double *result)
{
  DCCMdElement3x3 *DE;
  int j;

  assert(x);
  assert(result);
  
  for(j=0;j<3*dim3_;j++) result[j]=0.0;

  for(j=0;j<dim3_;j++)
    {
      DE=col[j];

      while(DE) 
	{
	  //result[DE->row]+=DE->val*x[j];
	  //for(int k=0;k<3;k++)
	  //for(int l=0;l<3;l++)
	  //result[3*DE->row+k]+=DE->val[3*k+l]*x[3*j+l];

	  int row;
	  row=DE->row;

	  result[3*row+0]+=DE->val[0]*x[3*j+0];
	  result[3*row+0]+=DE->val[1]*x[3*j+1];
	  result[3*row+0]+=DE->val[2]*x[3*j+2];
	  result[3*row+1]+=DE->val[3]*x[3*j+0];
	  result[3*row+1]+=DE->val[4]*x[3*j+1];
	  result[3*row+1]+=DE->val[5]*x[3*j+2];
	  result[3*row+2]+=DE->val[6]*x[3*j+0];
	  result[3*row+2]+=DE->val[7]*x[3*j+1];
	  result[3*row+2]+=DE->val[8]*x[3*j+2];
	  DE=DE->next;
	}
    }
}

//////////////////////////////////////////////////////////////////////////////

DCCMdElement3x3* DynamicCompCol_Mat_double3x3::find3x3(int i, int j) const
{
  DCCMdElement3x3 *DE;
  
  assert(0<=i);
  assert(i<dim3_);
  assert(0<=j);
  assert(j<dim3_);
  
  DE=col[j];
  
  while(DE && DE->row<=i)
    {
      if(DE->row==i) return(DE);
      DE=DE->next;
    }
  return(NULL);
}

//////////////////////////////////////////////////////////////////////////////

void DynamicCompCol_Mat_double3x3::add3x3(int i, int j, double *v)
{
  DCCMdElement3x3 *DE;
  
  //cout << "add " << i << ", " << j << ", " << v << endl;

  DE=find3x3(i,j);
  if(DE) 
    {
      DE->val[0]+=v[0];
      DE->val[1]+=v[1];
      DE->val[2]+=v[2];
      DE->val[3]+=v[3];
      DE->val[4]+=v[4];
      DE->val[5]+=v[5];
      DE->val[6]+=v[6];
      DE->val[7]+=v[7];
      DE->val[8]+=v[8];
    }
  else   set3x3(i,j,v);
}

//////////////////////////////////////////////////////////////////////////////

void DynamicCompCol_Mat_double3x3::set3x3(int i, int j, double *v)
{
  DCCMdElement3x3 *DE;
  
  DE=find3x3(i,j);
  if(DE) 
    {
      DE->val[0]=v[0];
      DE->val[1]=v[1];
      DE->val[2]=v[2];
      DE->val[3]=v[3];
      DE->val[4]=v[4];
      DE->val[5]=v[5];
      DE->val[6]=v[6];
      DE->val[7]=v[7];
      DE->val[8]=v[8];
    }
  else
    {
      DCCMdElement3x3 *old;
	  
      NumNonZeros_++;
      DE=col[j];

      if(!DE)
	{ // first element to insert in column
	  //cout << "first element to insert" << endl;
	  col[j]=new DCCMdElement3x3(NULL,i,v);
	  return;
	}
      else
	{
	  while(DE)
	    {
	      //cout << "A" << endl;
	      if(DE->row<i) 
		{
		  old=DE;
		  DE=DE->next;
		  if(!DE)
		    { // insert at end of list
		      //cout << "insert at end of list" << endl;
		      old->next=new DCCMdElement3x3(NULL,i,v);
			  return;
		    }
		}
	      else
		{ // insert not at end of list
		  if(DE==col[j])
		    {
		      //cout << "insert as first element" << endl;
		      col[j]=new DCCMdElement3x3(DE,i,v);
		      //cout << "done" << endl;
		      return;
		    }
		  else
		    {
		      //cout << "insert in middle of list" << endl;
		      old->next=new DCCMdElement3x3(DE,i,v);
		      //cout << "done" << endl;
		      return;
		    }
		}
	    }
	}
    }
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
#ifdef UNDEFINED
void DynamicCompCol_Mat_double::Cholesky(void)
{
  int i,j,k;
  double a,c;

  for(i=0;i<dim();i++)
    {
      a=(*this)(i,i);
      if(a<=0.0) 
	{
	  cerr << "Cholesky: Matrix is not positive definite!"<<endl;
	  exit(1);
	}
      c=sqrt(a);
      set(i,i,c);

      DCCMdElement *DE1, *DE2;

      DE1=find(i,i);
      if(!DE1)
	{
	  cerr << "Cholesky: No diagonal element set (=0)!" <<endl;
	  exit(1);
	}
      DE2=DE1;

      // update v vector
      while(DE1->next)
	{
	  DE1=DE1->next;
	  DE1->val=DE1->val/c;
	}

      // update rest of matrix
      DE1=DE2; // back to (i,i)
      while(DE1)
	{
	  DE1=DE1->next;
	  DE2=DE1;
	  while(DE2)
	    {
	      double tmp;

	      tmp=-(DE1->val)*(DE2->val);

	      if(fabs(tmp)!=0.0)
		{
		  j=DE1->row;
		  k=DE2->row;

		  if(!find(k,j))  // k>=j
		    {
		      //cout << "Updating zero element!" << endl;
		    }		  
		  add(k,j,tmp);
		}

	      DE2=DE2->next;
	    }
	}
    }

  for(i=0;i<dim();i++)
    {
      DCCMdElement *DE;

      while(col[i] && col[i]->row<i)
	{
	  DE=col[i];
	  col[i]=DE->next;
	  delete DE;
	}
    }
  cout << "Finished Cholesky decomposition." << endl;
  clear_from_zeros();
}
#endif
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
