#include "CCSubdivide.h"

Point::Point(FILE *f)
{
  fscanf(f,"%d %d %d %le %le %le\n",&ID,&constraint,&valence,&x,&y,&z);
  valence=0;
}

Quad::Quad(FILE *f)
{
  int i,j;
  fscanf(f,"%d %d %d %d\n",p,p+1,p+2,p+3);

  for(i=0;i<4;i++) NQ[i]=NULL;
  centerP =-1;
  edgeP[0]=-1;
  edgeP[1]=-1;
  edgeP[2]=-1;
  edgeP[3]=-1;
}

void Quad::set_neighbor(Quad *N, int p1, int p2, int npc)
{
  int j,flag;
  
  flag=0;
  for(j=0;j<4;j++)
    {
      if(p[j]==p1)
	{
	  if(p[(j+1)%4]==p2)
	    {
	      edgeP[j]=npc;
	      NQ[j]=N;
	      flag++;
	    }
	  else if(p[(j+3)%4]==p2)
	    {
	      edgeP[(j+3)%4]=npc;
	      NQ[(j+3)%4]=N;
	      flag++;
	    }
	  else
	    {
	      printf("Error: Data structure broken!\n");
	      exit(1);
	    }	    
	}
    }
  if(flag!=1)
    {
      printf("Error: Data structure broken!\n");
      exit(1);
    }
}

Mesh::Mesh(FILE *f)
{
  int i;

  fscanf(f,"%d\n",&numpatches);
  printf("%d\n",numpatches);
  assert(0<numpatches);
  assert(numpatches <10000);
  Q=new Quad*[numpatches];
  for(i=0;i<numpatches;i++) Q[i]=new Quad(f);

  fscanf(f,"%d",&numpoints);
  assert(0<numpoints);
  assert(numpoints <10000);
  P=new Point*[numpoints];
  for(i=0;i<numpoints;i++) P[i]=new Point(f);

  NP=new Point*[9*numpatches];
  for(i=0;i<9*numpatches;i++) NP[i]=new Point();
  npc=0;

  set_valences();
  set_neighbors();
  //set_new_points();
  //subdivide();
}

void Mesh::set_valences(void)
{
  int i,j;

  for(i=0;i<numpatches;i++)
    {
      for(j=0;j<4;j++)
	{
	  P[Q[i]->p[j]]->adjacent[P[Q[i]->p[j]]->valence]=i;
	  P[Q[i]->p[j]]->valence++;

	  if(P[Q[i]->p[j]]->valence==MAXVALENCE)
	    {
	      printf("Error: Increase MAXVALENCE!\n");
	      exit(1);
	    }
	}
    }

  for(i=0;i<numpoints;i++)
    {
      double w;

      w=1.0-1.5/P[i]->valence-0.25/P[i]->valence;

      NP[i]->x=w*P[i]->x;
      NP[i]->y=w*P[i]->y;
      NP[i]->z=w*P[i]->z;
    }
    
  for(i=0;i<numpatches;i++)
    {
      for(j=0;j<4;j++)
	{
	  NP[Q[i]->p[j]]->x+=0.75 /pow(P[Q[i]->p[j]]->valence,2.0)*(P[Q[i]->p[(j+1)%4]]->x+
								   P[Q[i]->p[(j+3)%4]]->x)+
  	                     0.25/pow(P[Q[i]->p[j]]->valence,2.0)* P[Q[i]->p[(j+2)%4]]->x;
	  NP[Q[i]->p[j]]->y+=0.75 /pow(P[Q[i]->p[j]]->valence,2.0)*(P[Q[i]->p[(j+1)%4]]->y+
								   P[Q[i]->p[(j+3)%4]]->y)+
  	                     0.25/pow(P[Q[i]->p[j]]->valence,2.0)* P[Q[i]->p[(j+2)%4]]->y;
	  NP[Q[i]->p[j]]->z+=0.75 /pow(P[Q[i]->p[j]]->valence,2.0)*(P[Q[i]->p[(j+1)%4]]->z+
								   P[Q[i]->p[(j+3)%4]]->z)+
  	                     0.25/pow(P[Q[i]->p[j]]->valence,2.0)* P[Q[i]->p[(j+2)%4]]->z;
	}
    }
}

void Mesh::set_neighbors(void)
{
  int i,j,k;

  npc=numpoints;
  for(i=0;i<numpatches;i++) // all patches
    {
      Q[i]->centerP=npc;
      NP[npc]->x = 0.25*(P[Q[i]->p[0]]->x +
			 P[Q[i]->p[1]]->x +
			 P[Q[i]->p[2]]->x +
			 P[Q[i]->p[3]]->x);
      NP[npc]->y = 0.25*(P[Q[i]->p[0]]->y +
			 P[Q[i]->p[1]]->y +
			 P[Q[i]->p[2]]->y +
			 P[Q[i]->p[3]]->y);
      NP[npc]->z = 0.25*(P[Q[i]->p[0]]->z +
			 P[Q[i]->p[1]]->z +
			 P[Q[i]->p[2]]->z +
			 P[Q[i]->p[3]]->z);
      npc++;
    }

  for(i=0;i<numpatches;i++) // all patches
    {
      for(j=0;j<4;j++) // all points/links in patch
	{
	  int flag;
	  flag=0;

	  if(!Q[i]->NQ[j])
	    {
	      for(k=0;k<P[Q[i]->p[j]]->valence;k++) // all adjacent quads
		{
		  if(Q[P[Q[i]->p[j]]->adjacent[k]]->has_point(Q[i]->p[(j+1)%4]))
		    {
		      Q[i]->NQ[j]=Q[P[Q[i]->p[j]]->adjacent[k]];
		      Q[i]->edgeP[j]=npc;
		      Q[i]->NQ[j]->set_neighbor(Q[i],Q[i]->p[j],Q[i]->p[(j+1)%4],npc);

		      npc++;
		      flag=1;
		    }
		}
	    }
	  //if(!flag) printf("No link in %d %d\n",i,j);
	}
    }

  for(i=0;i<numpatches;i++)
    {
      for(j=0;j<4;j++)
	{
	  if(!Q[i]->NQ[j]) printf("Warning: patch %d link %d null.\n",i,j);
	}
    }

   for(i=0;i<numpatches;i++) // all patches
    {
      for(j=0;j<4;j++)
	{
	  NP[Q[i]->edgeP[j]]->x += 3.0/16.0*(P[Q[i]->p[ j     ]]->x +
					     P[Q[i]->p[(j+1)%4]]->x)+
	                           1.0/16.0*(P[Q[i]->p[(j+2)%4]]->x +
					     P[Q[i]->p[(j+3)%4]]->x);
	  NP[Q[i]->edgeP[j]]->y += 3.0/16.0*(P[Q[i]->p[ j     ]]->y +
					     P[Q[i]->p[(j+1)%4]]->y)+
	                           1.0/16.0*(P[Q[i]->p[(j+2)%4]]->y +
					     P[Q[i]->p[(j+3)%4]]->y);
	  NP[Q[i]->edgeP[j]]->z += 3.0/16.0*(P[Q[i]->p[ j     ]]->z +
					     P[Q[i]->p[(j+1)%4]]->z)+
	                           1.0/16.0*(P[Q[i]->p[(j+2)%4]]->z +
					     P[Q[i]->p[(j+3)%4]]->z);
	}
    }
}
/*
void Mesh::set_new_points(void)
{
  int i,j,k;
  new_numpoints=0;
  NP=new Point*[9*numpatches];
  
  for(i=0;i<numpatches;i++)
    {
      for(j=0;j<3;j++)
	{
	  for(k=0;k<3;k++)
	    {
	      if();// *****************
	    }
	}
    }
}
*/
void Mesh::save(char *fname)
{
  FILE *f;
  int i;

  f=fopen(fname,"w");

  fprintf(f,"%d\n",4*numpatches);
  for(i=0;i<numpatches;i++)
    {
      Q[i]->save(f);
    }
  fprintf(f,"%d\n",npc);
  for(i=0;i<npc;i++)
    {
      fprintf(f,"%d %e %e %e\n",i,
	      NP[i]->x,
	      NP[i]->y,
	      NP[i]->z);
    }
  fclose(f);
}


int main(int argc, char **argv)
{
  if(argc!=3)
    {
      printf("%s infile outfile\n",argv[0]);
      exit(1);
    }

  FILE *f;

  f=fopen(argv[1],"r");
  Mesh M(f);
  fclose(f);
  M.save(argv[2]);
}
