//////////////////////////////////////////////////////////////////////////////
//
// 
//
//
//////////////////////////////////////////////////////////////////////////////


#include "Minimizer.h"

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//
Linear_Minimizer::~Linear_Minimizer()
{

}

//////////////////////////////////////////////////////////////////////////////

Linear_Minimizer::Linear_Minimizer()
{

}

//////////////////////////////////////////////////////////////////////////////
/*
void Linear_Minimizer::LinearStep(void)
{
  DynamicCompCol_Mat_double *H;
  double                    *G;
  double tol = 1.e-6;                    // Convergence tolerance
  int result, maxit = 150;               // Maximum iterations

  H=new DynamicCompCol_Mat_double(3*numpoints);
  G=new double[3*numpoints];

  mesh->Hess_KpEnergy(H);
  mesh->grad_KpEnergy(G);
  //neg(G);

  CompCol_Mat_double A(3*numpoints, 3*numpoints,
		       H.NumNonZeros(), 
		       H.val(),
		       H.row_ind(),
		       H.col_ptr());

  VECTOR_double b(G, 3*numpoints);
  VECTOR_double x(3*numpoints, 0.0);     // initial guess

  DiagPreconditioner_double D(A);        // Create diagonal preconditioner
  result = CG(A, x, b, D, maxit, tol);   // solve system

  (*P_new)=(*P_old)-x;
}
*/
//////////////////////////////////////////////////////////////////////////////

void Linear_Minimizer::minimize(Mesh *M)
{
  double E_old, E_new;
  DynamicCompCol_Mat_double *H;//,*H2;
  double                    *G;
  int i, iteration;
  char fname[256];

  mesh      = M;
  numpoints = M->numpoints;

  P_old=new double[3*numpoints];
  P_new=new double[3*numpoints];
      G=new double[3*numpoints];

  copy(M->P, P_old);

  E_old=M->KpEnergy(P_old);
  E_new=E_old;

  //cout << "Exact bending energy = ";
  //fflush(0);
  //cout << M->Energy(P_old) << endl;
  //M->compare_energies(P_old);

  iteration=0;
  //while(E_old>=E_new)
    {
      double tol = 1.e-12;                  // Convergence tolerance
      int result, maxit = 10000;              // Maximum iterations

      iteration++;
      H=mesh->Hess_KpEnergy(P_old);
      //cout << "Hess finished" << endl;
      //fflush(0);

      mesh->grad_KpEnergy(P_old, G);
      //cout << "Grad finished" << endl;
      //fflush(0);
      
#ifdef DEBUG
      cout << "Saving matrix as ppm."<<endl;
      sprintf(fname,"matrix%d.ppm",iteration);
      H->save(fname);      
#endif
      CompCol_Mat_double A(3*numpoints, 3*numpoints,
			   H->NumNonZeros(), 
			   H->val(),
			   H->row_ind(),
			   H->col_ptr());

#ifdef DEBUG
      //for(i=0;i<3*numpoints;i++)
      //{
      //cout << G[i] << endl;
      //}

      cout << "Checking for symmetry" << endl; fflush(0);
      for(i=0;i<3*numpoints;i++)
	{
	  int j;
	  for(j=0;j<3*numpoints;j++)
	    {
	      if(fabs((*H)(i,j)-(*H)(j,i))>1e-6) 
		cout << i << ", " << j << ", " << (*H)(i,j)<<"!="<< (*H)(j,i) << endl;
	    }
	  //cout << endl;
	}
#endif

      //cout << "Computing the Cholesky decomposition." << endl;
      //H->Cholesky();
      //H->save("Cholesky.ppm");

      // test, if Cholesky gives the right result
      /*
      H2=mesh->Hess_KpEnergy(P_old);      int i,j,k;
      for(i=0;i<H->dim();i++)
	{
	  cout <<" "<<i; fflush(0);
	  for(j=0;j<H->dim();j++)
	    {
	      double tmp;

	      tmp=0.0;
	      for(k=0;k<H->dim();k++)
		{
		  tmp+=(*H)(i,k)*(*H)(j,k);
		}
	      if(fabs((*H2)(i,j)-tmp)>1e-6)
		{
		  cerr << "Cholesky decomp problem!" << endl;
		  cerr << "at i="<<i<<", j="<<j<< endl;
		  cerr << "tmp="<<tmp<<", H2="<<(*H2)(i,j) << endl;
		  exit(1);
		}
	    }
	}
      */
      /*
      for(i=3*numpoints-30;i<3*numpoints;i++)
      {
	cout << G[i] << endl;
      }
      cout << "done" << endl; fflush(0);
      */
      VECTOR_double b(G, 3*numpoints);
      VECTOR_double x(3*numpoints, 0.0);     // initial guess
      
      DiagPreconditioner_double D(A);        // Create diagonal preconditioner

      cout << "Solving equation system." << endl; fflush(0);
      result = BiCG(A, x, b, D, maxit, tol);   // solve system
      cout << "Iterations " << maxit << " ("<<result <<")"<< endl;
      cout << "finished..." << endl; fflush(0);

      // compute new point vector
      for(i=0;i<3*numpoints;i++)
      {
	P_new[i]=P_old[i]-x(i);
      }


      x=A*b;
      for(i=3*numpoints-0;i<3*numpoints;i++)
      {
	cout << P_new[i] << ", " << x(i) << ", " << b(i) << endl;
      }

      E_old=E_new;
      E_new=M->KpEnergy(P_new);

      printf("E_old=%1.17e\nE_new=%1.17e\n",E_old,E_new);

      P_tmp=P_old;
      P_old=P_new;
      P_new=P_old;

      sprintf(fname,"min%d.iv",iteration);

      for(i=0;i<3*numpoints;i++)
	{
	  G[i]=0.0;
	}
      //mesh->grad_KpEnergy(P_old,G);
      //mesh->export_IV(fname,P_old,G);
      //sprintf(fname,"min%d",iteration);
      //mesh->save(fname);

      delete H;
    }

  //cout << "Exact bending energy = ";
  //fflush(0);
  //cout << M->Energy(P_new) << endl;

  copy(P_new,M->P);    // write solution into Mesh
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
