//////////////////////////////////////////////////////////////////////////////
//
// PatchTest.C, 06/28/01, Ilja Friedel
//
//////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <math.h>
#include <iostream.h>
#include <stdlib.h>

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

#include "Patch/Patch.h"
#include "Patch/UnitPatch.h"
#include "Patch/BilinearPatch.h"
#include "Patch/BicubicPatch.h"
#include "Patch/EigenStructure.h"
#include "Patch/ExtraordinaryPatch.h"
#include "Patch/EigenPatch.h"
#include "Patch/FirstOrderDataDependentPatch.h"
#include "CC_Mesh/Mesh.h"
#include "Minimizer/Minimizer.h"
 
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

int load_control_points(FILE *f, double **p)
{
  int  i,j,valence,M;
      
  fscanf(f,"%d",&valence);

  M=2*valence+8;

  for(i=0;i<M;i++)
    {
      for(j=0;j<3;j++)
	{
	  fscanf(f,"%lf",&(p[j][i]));
	}
    }
  return(valence);
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

int main(int argc, char ** argv)
{
  int    i;
  /*
  double tmp1, tmp2;
  int    N=61;
  int    M=8+2*N;
  double K[M*M];
  double K2[M*M];

  BilinearPatch      *BP;
  UnitPatch          *UP;
  ExtraordinaryPatch *EP;
  //BicubicPatch       *P ;
  EigenPatch         *EiP;
  FirstOrderDataDependentPatch *FODDP;

  p[0]=new double[M+2];
  p[1]=new double[M+2];
  p[2]=new double[M+2];
  
  FILE *f;
  */
  char fname[1024];

  if(argc!=2)
    {
      cout << argv[0] << " file" << endl;
      exit(1);
    }

  sprintf(fname,"%s",argv[1]);

  ///home/ilja/smoothing/code/curvature/cube_points/debug.mtx");

  Mesh *ME;
  MeshPatch *MP;
  //  FirstOrderDataDependentPatch *P;
  Minimizer *Min;

  ME=new Mesh(fname);
  assert(ME);

  double *G1;
  double *G2;
  
  G1=new double[1000000];
  G2=new double[1000000];
  assert(G1);
  assert(G2);

  MP=ME->firstpatch;  
  /*
  // coloring of relative error
  double SBendE, SMembE, SKpE, SKmE;
  SBendE=0.0;
  SMembE=0.0;
  SKpE  =0.0;
  SKmE  =0.0;
  while(MP)
    {
      double red, green, blue;
      double BendE, MembE, KpE, KmE;

      red  =0.2;
      green=0.2;
      blue =0.2;

      P=MP->create_Patch(ME->P);
      
      KpE   =P->KpEnergy();
      BendE =P->ExactBendingEnergy();
      KmE   =P->KmEnergy();
      MembE =P->ExactMembraneEnergy();

      if(BendE<EPSILON1)
	{
	  red  =0.2;
	  green=0.2;
	  blue =0.2;
	}
      else
	{
	  red=0.2*KpE/BendE;
	  if(red<0.2)
	    {
	      blue=0.04/red;
	      red=0.2;
	    }
	}
      P->export_IV("bending.iv",red,green,blue);

      red  =0.2;
      green=0.2;
      blue =0.2;

      if(MembE<EPSILON1)
	{
	  red  =0.2;
	  green=0.2;
	  blue =0.2;
	}
      else
	{
	  red=0.2*KmE/MembE;
	  if(red<0.2)
	    {
	      blue=0.04/red;
	      red=0.2;
	    }
	}
      P->export_IV("membrane.iv",red,green,blue);

      MP=MP->next;

      //cout << KpE/BendE<<"("<<BendE<<")"<<" ";fflush(0);
      SBendE+=BendE;
      SMembE+=MembE;
      SKpE  +=KpE;
      SKmE  +=KmE;
    }
  
  //printf("\n\n");
  //printf("Membrane Energy=%1.18e\n",SMembE);
  //printf("Km       Energy=%1.18e\n",SKmE);
  //printf("Bending  Energy=%1.18e\n",SBendE);
  //printf("Kp       Energy=%1.18e\n",SKpE);

  cerr << SMembE << "  " << SKmE << "  " << fabs((SMembE-SKmE)/SMembE) << "  "
       << SBendE << "  " << SKpE << "  " << fabs((SBendE-SKpE)/SBendE) << endl;
  */
  /*
  // create all matrices
  for(i=3;i<61;i++)
    {
      P=new FirstOrderDataDependentPatch(p,i);
      P->KpEnergy();
      delete P;     
    }
  */
  /* 
  // time measurements
  int loop=100;

  fflush(0);
  system("date");
  fflush(0);
  for(i=0;i<loop;i++)
    {
      ME->KpEnergy(ME->P);
    }
  fflush(0);
  system("date");
  fflush(0);
  for(i=0;i<loop;i++)
    {
      ME->grad_KpEnergy(ME->P,G1);
    }
  fflush(0);
  system("date");
  fflush(0);
  for(i=0;i<loop;i++)
    {
      printf("*");fflush(0);
      delete ME->Hess_KpEnergy(ME->P);
    }
  fflush(0);
  system("date");
  fflush(0);
  */

  /*
  // analyzing Hessian and comparing different methods
  MP=ME->firstpatch;  
  while(MP)
    {
      P=MP->create_Patch(ME->P);
      P->Hess_KpEnergy_central_diff(G1);
      P->Hess_KpEnergy(G2);

      cout << P->Valence << endl;
      for(i=0;i<2*P->Valence+8;i++)
	for(j=0;j<2*P->Valence+8;j++)
	  for(k=0;k<3;k++)
	    for(l=0;l<3;l++)
	      {
		//printf("Ex  %1.12e    %d %d  %d %d\n",G1[3*(2*P->Valence+8)*(3*i+k)+(3*j+l)],i,j,k,l);
		//printf("Kp  %1.12e        ",G2[3*(2*P->Valence+8)*(3*i+k)+(3*j+l)]);

		if(fabs(G1[3*(2*P->Valence+8)*(3*i+k)+(3*j+l)]-
			G2[3*(2*P->Valence+8)*(3*i+k)+(3*j+l)])>1e-4)
		  {
		    printf("\nEx  %1.12e    %d %d  %d %d\n",G1[3*(2*P->Valence+8)*(3*i+k)+(3*j+l)],i,j,k,l);
		    printf("Kp  %1.12e        ",G2[3*(2*P->Valence+8)*(3*i+k)+(3*j+l)]);
		    printf("*****\n");
		  }
		//printf("\n");

	      }
      
      MP=MP->next;
      }
  */
  /*
  // analyzing gradient and comparing different methods
  MP=ME->firstpatch;

  while(MP)
    {
      P=MP->create_Patch(ME->P);

      P->grad_KpEnergy(G2);

      cout << P->Valence << endl;
      P->grad_KpEnergy_central_diff(G1);
      P->grad_KpEnergy(G2);
      
      for(i=0;i<2*P->Valence+8;i++)
	{
	    printf("Ex  %1.12e  %1.12e  %1.12e\n",    G1[3*i+0],G1[3*i+1],G1[3*i+2]);
	    printf("Kp  %1.12e  %1.12e  %1.12e\n",    G2[3*i+0],G2[3*i+1],G2[3*i+2]);
	    
	    for(k=0;k<3;k++)
	      {
		if(fabs(G1[3*i+k]-G2[3*i+k])>1e-6)
		  {
		    printf("Ex  %1.12e    %d %d \n",G1[3*i+k],i,k);
		    printf("Kp  %1.12e        ",G2[3*i+k]);
		    printf("*****\n");
		  }
	      }
	  }
      MP=MP->next;
    }
  */
  
  //for(i=0;i<ME->numpoints*3;i++) G1[i]=0.0;
  //ME->grad_KpEnergy(ME->P,G1);
  //ME->export_IV("before.iv",ME->P,G1);

  //Min=new Newton_Minimizer();
  Min=new Linear_Minimizer();
  Min->minimize(ME);

  //Min->gradEnergy(ME->P,G1);
  //ME->export_IV("after.iv",ME->P,G1);
  sprintf(fname,"%s.linear",argv[1]);
  ME->save(fname);
  fflush(0);

  delete ME;
  return(1);
}
