//////////////////////////////////////////////////////////////////////////////
//
// 
//
//
//////////////////////////////////////////////////////////////////////////////

#include "Minimizer.h"

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//
CG_Minimizer::~CG_Minimizer()
{

}

//////////////////////////////////////////////////////////////////////////////

CG_Minimizer::CG_Minimizer()
{
  A   = NULL;
  B   = NULL;
  G   = NULL;
  H   = NULL;
  DEP = NULL;
  DEA = NULL;
  DEB = NULL;
  tmp = NULL;
  tmp2= NULL;
}

//////////////////////////////////////////////////////////////////////////////

void CG_Minimizer::minimize(Mesh *M)
{
  mesh     =M;
  numpoints=M->numpoints;
  conjugate_gradient_minimize();
}

//////////////////////////////////////////////////////////////////////////////
// algorithm based on "Numerical Recipes in C" chapter 10.2

double CG_Minimizer::brent(double *MP, double *D, 
		   double ax, double bx, double cx, double tol)
{
  int iteration;
  double a,b,d,etemp,fu,fv,fw,fx,p,q,r,tol1,tol2,u,v,w,x,xm;
  double e=0.0;

  a=(ax < cx ? ax:cx);
  b=(ax > cx ? ax:cx);

  x=w=v=bx;
  fw=fv=fx=line_Energy(MP,D,x);

  for(iteration=1;iteration<ITMAX;iteration++)
    {
      cout << "[" << iteration << "]";

      xm=0.5*(a+b);
      tol1 = tol*fabs(x)+ZEPS;
      tol2 = 2.0*tol1;

      if(fabs(x-xm) <= (tol2-0.5*(b-a)))
	{
	  return x;
	}
      if(fabs(e) > tol1)
	{
	  r=(x-w)*(fx-fv);
	  q=(x-v)*(fx-fw);
	  p=(x-v)*q-(x-w)*r;
	  q=2.0*(q-r);

	  if(q>0.0) p=-p;
	  q=fabs(q);

	  etemp=e;
	  e=d;
	  if(fabs(p)>=fabs(0.5*q*etemp) || 
	     p <= q*(a-x) ||
	     p >= q*(b-x))
	    {
	      e=(x >= xm ? a-x : b-x);
	      d=CGOLD*e;
	    }
	  else
	    {
	      d=p/q;
	      u=x+d;
	      if(u-a < tol2 || b-u < tol2)
		{
		  d=SIGN(tol1,xm-x);
		}	      
	    }
	}
      else
	{
	  e=(x>=xm ? a-x : b-x);
	  d=CGOLD*e;
	}
      u=(fabs(d) >= tol1 ? x+d : x+SIGN(tol1,d));
      fu=line_Energy(MP,D,u);

      if(fu<=fx)
	{
	  if(u>=x) a=x;
	  else     b=x;
	  SHFT(v,w,x,u);
	  SHFT(fv,fw,fx,fu);
	}
      else
	{
	  if(u<x) a=u;
	  else    b=u;
	  if(fu<=fw || w==x)
	    {
	      v=w;
	      w=u;
	      fv=fw;
	      fw=fu;		
	    }
	  else if(fu <= fv || v==x || v==w)
	    {
	      v=u;
	      fv=fu;
	    }
	}
    }
  cerr << "Too many iterations in brent" << endl;
  return 0.0;
}

//////////////////////////////////////////////////////////////////////////////
// algorithm based on "Numerical Recipes in C" chapter 10.3

double CG_Minimizer::dbrent(double *MP, double *D, 
		    double ax, double bx, double cx, double tol)
{
  int    iteration, ok1, ok2;
  double a,b,d,d1,d2,du,dv,dw,dx,e=0.0;
  double fu,fv,fw,fx,olde,tol1,tol2,u,u1,u2,v,w,x,xm;

  a=(ax < cx ? ax:cx);
  b=(ax > cx ? ax:cx);

  x=w=v=bx;
  fw=fv=fx=     line_Energy(MP,D,x);
  dw=dv=dx=diff_line_Energy(MP,D,x);

  for(iteration=1;iteration<=ITMAX;iteration++)
    {
      cout << "[" << iteration << "]";

      xm  = 0.5*(a+b);
      tol1= tol*fabs(x)+ZEPS;
      tol2= 2.0*tol1;

      if(fabs(x-xm) <= (tol2-0.5*(b-a)))
	{
	  return(x);
	}

      if(fabs(e) > tol1)
	{
	  d1=2.0*(b-a);
	  d2=d1;

	  if(dw!=dx)  d1=(w-x)*dx/(dx-dw);
	  if(dv!=dx)  d2=(v-x)*dx/(dx-dv);

	  u1=x+d1;
	  u2=x+d2;

	  ok1 = (a-u1)*(u1-b) > 0.0 && dx*d1 <= 0.0;
	  ok2 = (a-u2)*(u2-b) > 0.0 && dx*d2 <= 0.0;

	  olde=e;
	  e=d;

	  if(ok1 || ok2)
	    {
	      if(ok1 && ok2)
		{
		  d=(fabs(d1) < fabs(d2) ? d1 : d2);
		}
	      else if(ok1)
		{
		  d=d1;
		}
	      else
		{
		  d=d2;
		}
	      if(fabs(d) <= fabs(0.5*olde))
		{
		  u=x+d;
		  if(u-a < tol2 || b-u < tol2)
		    {
		      d=SIGN(tol1,xm-x);
		    }
		}
	      else
		{
		  e=(dx>=0.0 ? a-x : b-x);
		  d=0.5*e;
		}
	    }
	  else
	    {
	      e=(dx>=0.0 ? a-x : b-x);
	      d=0.5*e;
	    }
	}
      else
	{
	  e=(dx>=0.0 ? a-x : b-x);
	  d=0.5*e;
	}
      if(fabs(d)>=tol1)
	{
	  u=x+d;
	  fu=line_Energy(MP,D,u);
	}
      else
	{
	  u=x+SIGN(tol1,d);
	  fu=line_Energy(MP,D,u);
	  if(fu>fx)
	    {
	      return x;
	    }	      
	}
      du=diff_line_Energy(MP,D,u);
      if(fu<=fx)
	{
	  if(u>=x) a=x;
	  else     b=x;
	      
	  MOV3(v,fv,dv, w,fw,dw);
	  MOV3(w,fw,dw, x,fx,dx);
	  MOV3(x,fx,dx, u,fu,du);
	}
      else
	{
	  if(u<x) a=u;
	  else    b=u;
	  if(fu <= fw || w ==x)
	    {
	      MOV3(v,fv,dv, w,fw,dw);
	      MOV3(w,fw,dw, u,fu,du);
	    }
	  else if(fu < fv || v==x || v==w)
	    {
	      MOV3(v,fv,dv, u,fu,du);
	    }
	}
    }
  cerr << "Too many iterations in routine dbrent" << endl;
  return(0.0);
}

//////////////////////////////////////////////////////////////////////////////
// find brackets for the minimum and call brent

double CG_Minimizer::line_minimize(double *MP, double *D)
{
  double st;
  double a,b,c;
  double fa,fb,fc;
  
  st=0.0;
  a=b=c=st;
  cout << "--> line_minimize"<<endl;
  fa=fb=fc=line_Energy(MP,D,a);

  /* 
  while(fb>=fc)  // find right bracket
    {
      cout << ">";
  fflush(0);
  cout << "fb="<<fb<<", fc="<<fc<<endl;
  cout << " b="<<b<< ",  c="<<c<<endl;
  fflush(0);
      b  = c;
      fb = fc;
      c  = c+1e-12+fabs(c)*4.0;  
      fc = line_Energy(MP,D,c);
    }

  while(fa<=fb)  // find left bracket
    {
      cout << "<";
  fflush(0);
      b  = a;
      fb = fa;
      a  = a-1e-12-fabs(a)*4.0; 
      fa = line_Energy(MP,D,a);
    }
  */

  a=-1.0;
  c= 1.0;

  cout << endl;
  fflush(0);
  cout << "Interval  " << a << ", " << st << ", " << c << endl;
  cout << "Function  " << line_Energy(MP,D,a) 
       << ", " << line_Energy(MP,D,st) 
       << ", " << line_Energy(MP,D,c) << endl;

  return(brent(MP,D,a,st,c,1e-4)); // use original starting point
}

//////////////////////////////////////////////////////////////////////////////
// algorithm based on "Numerical Recipes in C" chapter 10.6
// "Conjugate Gradient Methods in Multidimensions"

void CG_Minimizer::conjugate_gradient_minimize(void)
{
  int i;
  char fname[1024];

  cout << "Before new"<<endl;

  A   = new double[3*numpoints];
  B   = new double[3*numpoints];
  G   = new double[3*numpoints];
  H   = new double[3*numpoints];
  DEP = new double[3*numpoints];
  DEA = new double[3*numpoints];
  DEB = new double[3*numpoints];
  tmp = new double[3*numpoints];
  tmp2= new double[3*numpoints];

  for(i=0;i<3*numpoints;i++)
    {
      A[i]   = 0.0;
      B[i]   = 0.0;
      G[i]   = 0.0;
      H[i]   = 0.0;
      DEP[i] = 0.0;
      DEA[i] = 0.0;
      DEB[i] = 0.0;
      tmp[i] = 0.0;
      tmp2[i]= 0.0;
    }

  cout << "After new"<<endl;
  /*
  grad_Energy(P,H);
  for(i=1;i<1000;i++)
    {
      double d,s;
      int j;
      s=line_minimize(P,H);// find the minimum on the line from P in direction H 
      scale_add(P,s,H,P);  // move points to line minimum

      grad_Energy(P,H);

      sprintf(fname,"min%d.iv",i);
      //if(!(i%2))
      export_IV(fname,P);
    }

  */
  mesh->grad_KpEnergy(mesh->P,G);        // G=grad(P)

  neg(G);                  // G=-G
  copy(G,H);               // H=G
  if(dot(G,G)==0.0)
    {
      cout << "Gradient zero. Already at minimum. Do nothing."<<endl;
      cout << "[void CG_Minimizer::conjugate_gradient_minimize(void)]"<<endl;
      return;
    }
  
  for(i=1;i<1000;i++)
    {
      double d,s;

      cout << "Line minimize..." << endl;
      s=line_minimize(mesh->P,H);// find the minimum on the line from P in direction H 
      scale_add(mesh->P,s,H,mesh->P);// move points to line minimum

      copy(G,tmp);         // tmp=G
      mesh->grad_KpEnergy(mesh->P,G);    // G=grad(P)  [compute P_i+1]

      // gamma_i = (g_i+1 - g_i)*g_i+1/(g_i*g_i) 
      d=dot(tmp,tmp);      // d=g_i*g_i
      if(d==0.0) return;
      sub(G,tmp,tmp);      // tmp=g_i+1 - g_i
      gamma=dot(tmp,G)/d;

      // h_i+1 = g_i+1 + gamma_i*h_i;
      scale(gamma, H);     // H=gamma*h_i
      add(G,H,H);          // h_i+1 = g_i+1 + H

      sprintf(fname,"min%d.iv",i);
      //if(!(i%2))
      mesh->export_IV(fname,mesh->P);
    }


  //delete[] P;    P   =NULL;
  delete[] A;    A   =NULL;
  delete[] B;    B   =NULL;
  delete[] G;    G   =NULL;
  delete[] H;    H   =NULL;
  delete[] DEP;  DEP =NULL;
  delete[] DEA;  DEA =NULL;
  delete[] DEB;  DEB =NULL;
  delete[] tmp;  tmp =NULL;
  delete[] tmp2; tmp2=NULL;
}

//////////////////////////////////////////////////////////////////////////////
// one dimensional function (in direction d)
double CG_Minimizer::line_Energy(double *MP, double *D, double s)
{
  double e;
  scale_add(MP,s,D,tmp);  // tmp=MP+s*D
  //cout << "CG_Minimizer::line_Energy after scale_add" << endl;
  e=mesh->KpEnergy(tmp);  //*************
  //cout << "CG_Minimizer::line_Energy after Energy" << endl;
  return e;
}

//////////////////////////////////////////////////////////////////////////////
// derivative of one dimensional function (in direction D)
double CG_Minimizer::diff_line_Energy(double *MP, double *D, double s)
{
  scale_add(MP,s,D,tmp);              // tmp=MP+s*D
  mesh->grad_KpEnergy(tmp,tmp2);              // tmp2=grad(E(tmp))
  return(dot(D,tmp2)/sqrt(dot(D,D))); // direction derivative in direction D
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

