//////////////////////////////////////////////////////////////////////////////
//
// multi_dim_minimize.cpp
//
//////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <math.h>
#include <iostream.h>
#include <stdlib.h>

//////////////////////////////////////////////////////////////////////////////

#define ITMAX 100
#define ZEPS  1.0e-15
#define CGOLD 0.3819660
#define MOV3(a,b,c,d,e,f) (a)=(d);(b)=(e);(c)=(f);
#define SIGN(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))
#define SHFT(a,b,c,d) (a)=(b);(b)=(c);(c)=(d);

//////////////////////////////////////////////////////////////////////////////

double f(double x)
{
  return(cos(x));
}

//////////////////////////////////////////////////////////////////////////////

double df(double x)
{
  return(-sin(x));
}

//////////////////////////////////////////////////////////////////////////////

double brent(double ax, double bx, double cx, double tol)
{
  int iteration;
  double a,b,d,etemp,fu,fv,fw,fx,p,q,r,tol1,tol2,u,v,w,x,xm;
  double e=0.0;

  a=(ax < cx ? ax:cx);
  b=(ax > cx ? ax:cx);

  x=w=v=bx;
  fw=fv=fx=f(x);

  for(iteration=1;iteration<ITMAX;iteration++)
    {
      cout << "Iteration " << iteration << endl;

      xm=0.5*(a+b);
      tol1 = tol*fabs(x)+ZEPS;
      tol2 = 2.0*tol1;

      if(fabs(x-xm) <= (tol2-0.5*(b-a)))
	{
	  return x;
	}
      if(fabs(e) > tol1)
	{
	  r=(x-w)*(fx-fv);
	  q=(x-v)*(fx-fw);
	  p=(x-v)*q-(x-w)*r;
	  q=2.0*(q-r);

	  if(q>0.0) p=-p;
	  q=fabs(q);

	  etemp=e;
	  e=d;
	  if(fabs(p)>=fabs(0.5*q*etemp) || 
	     p <= q*(a-x) ||
	     p >= q*(b-x))
	    {
	      e=(x >= xm ? a-x : b-x);
	      d=CGOLD*e;
	    }
	  else
	    {
	      d=p/q;
	      u=x+d;
	      if(u-a < tol2 || b-u < tol2)
		{
		  d=SIGN(tol1,xm-x);
		}	      
	    }
	}
      else
	{
	  e=(x>=xm ? a-x : b-x);
	  d=CGOLD*e;
	}
      u=(fabs(d) >= tol1 ? x+d : x+SIGN(tol1,d));
      fu=f(u);

      if(fu<=fx)
	{
	  if(u>=x) a=x;
	  else     b=x;
	  SHFT(v,w,x,u);
	  SHFT(fv,fw,fx,fu);
	}
      else
	{
	  if(u<x) a=u;
	  else    b=u;
	  if(fu<=fw || w==x)
	    {
	      v=w;
	      w=u;
	      fv=fw;
	      fw=fu;		
	    }
	  else if(fu <= fv || v==x || v==w)
	    {
	      v=u;
	      fv=fu;
	    }
	}
    }
  cerr << "Too many iterations in brent" << endl;
  return 0.0;
}

//////////////////////////////////////////////////////////////////////////////

double dbrent(double ax, double bx, double cx, double tol)
{
  int    iteration, ok1, ok2;
  double a,b,d,d1,d2,du,dv,dw,dx,e=0.0;
  double fu,fv,fw,fx,olde,tol1,tol2,u,u1,u2,v,w,x,xm;

  a=(ax < cx ? ax:cx);
  b=(ax > cx ? ax:cx);

  x=w=v=bx;
  fw=fv=fx= f(x);
  dw=dv=dx=df(x);

  for(iteration=1;iteration<=ITMAX;iteration++)
    {
      cout << "Iteration " << iteration << endl;

      xm  = 0.5*(a+b);
      tol1= tol*fabs(x)+ZEPS;
      tol2= 2.0*tol1;

      if(fabs(x-xm) <= (tol2-0.5*(b-a)))
	{
	  return(x);
	}

      if(fabs(e) > tol1)
	{
	  d1=2.0*(b-a);
	  d2=d1;

	  if(dw!=dx)  d1=(w-x)*dx/(dx-dw);
	  if(dv!=dx)  d2=(v-x)*dx/(dx-dv);

	  u1=x+d1;
	  u2=x+d2;

	  ok1 = (a-u1)*(u1-b) > 0.0 && dx*d1 <= 0.0;
	  ok2 = (a-u2)*(u2-b) > 0.0 && dx*d2 <= 0.0;

	  olde=e;
	  e=d;

	  if(ok1 || ok2)
	    {
	      if(ok1 && ok2)
		{
		  d=(fabs(d1) < fabs(d2) ? d1 : d2);
		}
	      else if(ok1)
		{
		  d=d1;
		}
	      else
		{
		  d=d2;
		}
	      if(fabs(d) <= fabs(0.5*olde))
		{
		  u=x+d;
		  if(u-a < tol2 || b-u < tol2)
		    {
		      d=SIGN(tol1,xm-x);
		    }
		}
	      else
		{
		  e=(dx>=0.0 ? a-x : b-x);
		  d=0.5*e;
		}
	    }
	  else
	    {
	      e=(dx>=0.0 ? a-x : b-x);
	      d=0.5*e;
	    }
	}
      else
	{
	  e=(dx>=0.0 ? a-x : b-x);
	  d=0.5*e;
	}
      if(fabs(d)>=tol1)
	{
	  u=x+d;
	  fu=f(u);
	}
      else
	{
	  u=x+SIGN(tol1,d);
	  fu=f(u);
	  if(fu>fx)
	    {
	      return x;
	    }	      
	}
      du=df(u);
      if(fu<=fx)
	{
	  if(u>=x) a=x;
	  else     b=x;
	      
	  MOV3(v,fv,dv, w,fw,dw);
	  MOV3(w,fw,dw, x,fx,dx);
	  MOV3(x,fx,dx, u,fu,du);
	}
      else
	{
	  if(u<x) a=u;
	  else    b=u;
	  if(fu <= fw || w ==x)
	    {
	      MOV3(v,fv,dv, w,fw,dw);
	      MOV3(w,fw,dw, u,fu,du);
	    }
	  else if(fu < fv || v==x || v==w)
	    {
	      MOV3(v,fv,dv, u,fu,du);
	    }
	}
    }
  cerr << "Too many iterations in routine dbrent" << endl;
  return(0.0);
}

//////////////////////////////////////////////////////////////////////////////

double minimize(double st)
{
  double a,b,c;
  double fa,fb,fc;
  double step;
  
  a=b=c=st;
  fa=fb=fc=f(a);

  while(fb>=fc)  // find right bracket
    {
      b  = c;
      fb = fc;
      c  = c+fabs(c)*1.2;  
      fc = f(c);
    }

  while(fa<=fb)  // find left bracket
    {
      b  = a;
      fb = fa;
      a  = a-fabs(a)*1.2; 
      fa = f(a);
    }

  cout << "Interval  " << a << ", " << st << ", " << c << endl;
  cout << "Function  " << f(a) << ", " << f(st) << ", " << f(c) << endl;

  return(brent(a,st,c,1e-10)); // use original starting point
}

//////////////////////////////////////////////////////////////////////////////

int main(int argc, char **argv)
{
  printf("Minimum at %1.15e\n", minimize(0.1));
  printf("Pi =       %1.15e\n", M_PI);
  return(1);
}

//////////////////////////////////////////////////////////////////////////////

