//////////////////////////////////////////////////////////////////////////////
//
// EigenStructure.h, 07/18/01, Ilja Friedel
//
//////////////////////////////////////////////////////////////////////////////

#ifndef EIGENSTRUCTURE_USED
#define EIGENSTRUCTURE_USED

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <cmath>
#include <stdlib.h>
#include <iostream>
using namespace std;
#include "BicubicPatch.h"

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

extern BicubicPatch **EigenBasis[MAX_VALENCE+1][3];

extern double *CJI[MAX_VALENCE+1];
extern double *CIJ[MAX_VALENCE+1];
extern double *CPI[MAX_VALENCE+1];
extern double *CIP[MAX_VALENCE+1];
extern double *iX [MAX_VALENCE+1]; // inverse eigenvectors
extern double *eX [MAX_VALENCE+1]; // eigenvectors

extern double *Lambda[MAX_VALENCE+1]; 
extern double *Omega1[MAX_VALENCE+1];
extern double *Omega2[MAX_VALENCE+1];
extern double *Omega3[MAX_VALENCE+1];
extern double *Omega4[MAX_VALENCE+1];

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

class EigenStructure
{
   
////////////////////////////////////////////////////////////////////////////// 

  static int initialized;

  Patch * EP;

  int     N;    // valence
  int     M;    // 2*N+8

  void   load_matrix(char *fname, double *p, int n, int m);
  void   save_matrix(char *fname, double *p, int n, int m);
  void   load_all_general_matrices(void);
  void   matrix_mult( double *A, double *B, double *C, int I, int J, int K );
  void   create_EigenFunctions();
  void   project_controlpoints_to_eigenbasis(void);

  double clip(double x)
    {
      if(fabs(x)<1e-10) return(0.0);
      else return(x);
    }

 public:
  double *eigen_controlpoints[3];

  double evaluate_phi    (double u, double v, int mode);
  double evaluate_D1phi  (double u, double v, int mode);
  double evaluate_D2phi  (double u, double v, int mode);

  double evaluate_D1D1phi(double u, double v, int mode);
  double evaluate_D1D2phi(double u, double v, int mode);
  double evaluate_D2D1phi(double u, double v, int mode);
  double evaluate_D2D2phi(double u, double v, int mode);

  ~EigenStructure();
  EigenStructure(Patch *EP);
};

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

#endif
