//////////////////////////////////////////////////////////////////////////////
// Patch.C
//
//
//////////////////////////////////////////////////////////////////////////////

#include "Patch.h"

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

#define REFLECTION_MAP

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

Patch::~Patch()
{
  // nothing to do
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

void Patch::export_IV(const char *fn,double r, double g, double b)
{
  int N,M;
  int i,j;
  FILE *f;

  N=SIZE_N;
  M=SIZE_M;

  f=fopen(fn,"a");
  assert(f);
  
  fprintf(f,"#Inventor V2.0 ascii\n\n");
  fprintf(f,"Separator {\n");
  fprintf(f,"	LightModel\n");
  fprintf(f,"	{\n");
  fprintf(f,"	   model BASE_COLOR\n");
  fprintf(f,"	}\n");
  fprintf(f,"	ShapeHints {\n");
  fprintf(f,"	shapeType UNKNOWN_SHAPE_TYPE\n");
  fprintf(f,"	vertexOrdering	COUNTERCLOCKWISE\n");
  fprintf(f,"	}\n");
  fprintf(f,"    PackedColor {\n");
  fprintf(f,"	rgba	[ \n");

  for(i=0;i<=N;i++)
    {
      for(j=0;j<=M;j++)
	{
	  unsigned int  color;
	  unsigned char red, green, blue;
	  
	  if(b<0.0) b=0.0;
	  if(g<0.0) g=0.0;
	  if(r<0.0) r=0.0;

	  if(b>1.0) b=1.0;
	  if(g>1.0) g=1.0;
	  if(r>1.0) r=1.0;

	  blue  = (int)(b*255.0);	  
	  green = (int)(g*255.0);
	  red   = (int)(r*255.0);

	  color=256*(256*(unsigned int)blue + (unsigned int)green) + red;
	      
	  if(i==N && j==M)
	    {
	      fprintf(f,"         0xff%.6x",color);
	    }
	  else
	    {
	      fprintf(f,"         0xff%.6x,",color);
	    }
	}
      fprintf(f,"\n");
    }

  fprintf(f,"]\n");
  fprintf(f,"    }\n");
  fprintf(f,"    MaterialBinding {\n");
  fprintf(f,"	value	PER_VERTEX_INDEXED\n");
  fprintf(f,"    }\n");
  fprintf(f,"    Coordinate3 {\n");
  fprintf(f,"	point	[\n");

  for(i=0;i<=N;i++)
    {
      for(j=0;j<=M;j++)
	{
	  if(i==N && j==M)
	    {
	      fprintf(f,"         %e %e %e",
		      evaluate_S(i/(double)N,j/(double)M, 0),
		      evaluate_S(i/(double)N,j/(double)M, 1),
		      evaluate_S(i/(double)N,j/(double)M, 2));
	    }
	  else
	    {
	      fprintf(f,"         %e %e %e,",
		      evaluate_S(i/(double)N,j/(double)M, 0),
		      evaluate_S(i/(double)N,j/(double)M, 1),
		      evaluate_S(i/(double)N,j/(double)M, 2));
	    }
	  fprintf(f,"\n");
	}
    }

  fprintf(f,"    ]}\n");
  fprintf(f,"    QuadMesh {\n");
  fprintf(f,"	verticesPerColumn	%d\n",N+1);
  fprintf(f,"	verticesPerRow	%d\n",M+1);
  fprintf(f,"    }\n");
  fprintf(f,"}\n");

  fclose(f);
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

void Patch::export_IV_gaussian_coloring(const char *fn, double scale)
{
  int N,M;
  int i,j;
  FILE *f;

  assert(scale>=0.0);
  N=SIZE_N;
  M=SIZE_M;

  f=fopen(fn,"a");
  assert(f);
  
  fprintf(f,"#Inventor V2.0 ascii\n\n");
  fprintf(f,"Separator {\n");
  fprintf(f,"	LightModel\n");
  fprintf(f,"	{\n");
  fprintf(f,"	   model BASE_COLOR\n");
  fprintf(f,"	}\n");
  fprintf(f,"	ShapeHints {\n");
  fprintf(f,"	shapeType UNKNOWN_SHAPE_TYPE\n");
  fprintf(f,"	vertexOrdering	COUNTERCLOCKWISE\n");
  fprintf(f,"	}\n");
  fprintf(f,"    PackedColor {\n");
  fprintf(f,"	rgba	[ \n");


  for(i=0;i<=N;i++)
    {
      for(j=0;j<=M;j++)
	{
	  unsigned int  color;
	  unsigned char red, green, blue;
	  double dR, dG, dB;
	  double lambda1;
	  
	  dR=0.0;
	  dG=0.0;
	  dB=0.0;
	      
	  if(i==0 && j==0)
	    {
	      lambda1=evaluate_Gaussian_curvature(1e-3,1e-3);
	    }
	  else 
	    {
	      lambda1=evaluate_Gaussian_curvature(i/(double)N,j/(double)M);
	    }

	  if(lambda1>=0.0)
	    {
	      if(lambda1<=1.0) dR  = lambda1;
	      else dR = 1.0;
	      dG  = 0.0;
	      dB  = 1.0-dR;
	    }
	  else
	    {
	      dR  = 0.0;
	      if(lambda1>=-1.0) dG  = -lambda1;
	      else dG = 1.0;
	      dB  = 1.0-dG;
	    }
	  
	  red   = (unsigned char)(255.0 * dR);
	  green = (unsigned char)(255.0 * dG);
	  blue  = (unsigned char)(255.0 * dB);
	  
	  color=256*(256*(unsigned int)blue + (unsigned int)green) + red;
	  
	  if(i==N && j==M)
	    {
	      fprintf(f,"         0xff%.6x",color);
	    }
	  else
	    {
	      fprintf(f,"         0xff%.6x,",color);
	    }
	}
      fprintf(f,"\n");
    }

  fprintf(f,"]\n");
  fprintf(f,"    }\n");
  fprintf(f,"    MaterialBinding {\n");
  fprintf(f,"	value	PER_VERTEX_INDEXED\n");
  fprintf(f,"    }\n");
  fprintf(f,"    Coordinate3 {\n");
  fprintf(f,"	point	[\n");

  for(i=0;i<=N;i++)
    {
      for(j=0;j<=M;j++)
	{
	  if(i==N && j==M)
	    {
	      fprintf(f,"         %e %e %e",
		      evaluate_S(i/(double)N,j/(double)M, 0),
		      evaluate_S(i/(double)N,j/(double)M, 1),
		      evaluate_S(i/(double)N,j/(double)M, 2));
	    }
	  else
	    {
	      fprintf(f,"         %e %e %e,",
		      evaluate_S(i/(double)N,j/(double)M, 0),
		      evaluate_S(i/(double)N,j/(double)M, 1),
		      evaluate_S(i/(double)N,j/(double)M, 2));
	    }
	  fprintf(f,"\n");
	}
    }

  fprintf(f,"    ]}\n");
  fprintf(f,"    QuadMesh {\n");
  fprintf(f,"	verticesPerColumn	%d\n",N+1);
  fprintf(f,"	verticesPerRow	%d\n",M+1);
  fprintf(f,"    }\n");
  fprintf(f,"}\n");

  fclose(f);
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

void Patch::export_IV_mean_coloring(const char *fn, double scale)
{
  int N,M;
  int i,j;
  FILE *f;

  assert(scale>0.0);
  N=SIZE_N;
  M=SIZE_M;

  f=fopen(fn,"a");
  assert(f);
  
  fprintf(f,"#Inventor V2.0 ascii\n\n");
  fprintf(f,"Separator {\n");
  fprintf(f,"	LightModel\n");
  fprintf(f,"	{\n");
  fprintf(f,"	   model BASE_COLOR\n");
  fprintf(f,"	}\n");
  fprintf(f,"	ShapeHints {\n");
  fprintf(f,"	shapeType UNKNOWN_SHAPE_TYPE\n");
  fprintf(f,"	vertexOrdering	COUNTERCLOCKWISE\n");
  fprintf(f,"	}\n");
  fprintf(f,"    PackedColor {\n");
  fprintf(f,"	rgba	[ \n");

  for(i=0;i<=N;i++)
    {
      for(j=0;j<=M;j++)
	{
	  unsigned int  color;
	  unsigned char red, green, blue;
	  double dR, dG, dB;
	  double lambda1;
	      
	  dR=0.0;
	  dG=0.0;
	  dB=0.0;
	  
	  if(i==0 && j==0)
	    {
	      lambda1=evaluate_mean_curvature(1e-3,1e-3);
	    }
	  else
	    {
	      lambda1=evaluate_mean_curvature(i/(double)N,j/(double)M);
	    }

	  if(lambda1>=0.0)
	    {
	      if(lambda1<=1.0) dR  = lambda1;
	      else dR = 1.0;
	      dG  = 0.0;
	      dB  = 1.0-dR;
	    }
	  else
	    {
	      dR  = 0.0;
	      if(lambda1>=-1.0) dG  = -lambda1;
	      else dG = 1.0;
	      dB  = 1.0-dG;
	    }
	  
	  red   = (unsigned char)(255.0 * dR);
	  green = (unsigned char)(255.0 * dG);
	  blue  = (unsigned char)(255.0 * dB);
	  
	  color=256*(256*(unsigned int)blue + (unsigned int)green) + red;
	  
	  if(i==N && j==M)
	    {
	      fprintf(f,"         0xff%.6x",color);
	    }
	  else
	    {
	      fprintf(f,"         0xff%.6x,",color);
	    }
	}
      fprintf(f,"\n");
    }

  fprintf(f,"]\n");
  fprintf(f,"    }\n");
  fprintf(f,"    MaterialBinding {\n");
  fprintf(f,"	value	PER_VERTEX_INDEXED\n");
  fprintf(f,"    }\n");
  fprintf(f,"    Coordinate3 {\n");
  fprintf(f,"	point	[\n");

  for(i=0;i<=N;i++)
    {
      for(j=0;j<=M;j++)
	{
	  if(i==N && j==M)
	    {
	      fprintf(f,"         %e %e %e",
		      evaluate_S(i/(double)N,j/(double)M, 0),
		      evaluate_S(i/(double)N,j/(double)M, 1),
		      evaluate_S(i/(double)N,j/(double)M, 2));
	    }
	  else
	    {
	      fprintf(f,"         %e %e %e,",
		      evaluate_S(i/(double)N,j/(double)M, 0),
		      evaluate_S(i/(double)N,j/(double)M, 1),
		      evaluate_S(i/(double)N,j/(double)M, 2));
	    }
	  fprintf(f,"\n");
	}
    }

  fprintf(f,"    ]}\n");
  fprintf(f,"    QuadMesh {\n");
  fprintf(f,"	verticesPerColumn	%d\n",N+1);
  fprintf(f,"	verticesPerRow	%d\n",M+1);
  fprintf(f,"    }\n");
  fprintf(f,"}\n");

  fclose(f);
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

void Patch::export_IV_principal_coloring(const char *fn,double scale)
{
  int N,M;
  int i,j;
  FILE *f;

  assert(scale>0.0);
  N=SIZE_N;
  M=SIZE_M;

  f=fopen(fn,"a");
  assert(f);
  
  fprintf(f,"#Inventor V2.0 ascii\n\n");
  fprintf(f,"Separator {\n");
  fprintf(f,"	LightModel\n");
  fprintf(f,"	{\n");
  fprintf(f,"	   model BASE_COLOR\n");
  fprintf(f,"	}\n");
  fprintf(f,"	ShapeHints {\n");
  fprintf(f,"	shapeType UNKNOWN_SHAPE_TYPE\n");
  fprintf(f,"	vertexOrdering	COUNTERCLOCKWISE\n");
  fprintf(f,"	}\n");
  fprintf(f,"    PackedColor {\n");
  fprintf(f,"	rgba	[ \n");

  for(i=0;i<=N;i++)
    {
      for(j=0;j<=M;j++)
	{
	  unsigned int  color;
	  unsigned char red, green, blue;
	  double dR, dG, dB;
	  double lambda1, lambda2;
	  
	  dR=0.0;
	  dG=0.0;
	  dB=0.0;
	  
	  if(i==0 && j==0)
	    {
	      set_principal_curvature(1e-3,1e-3);
	    }
	  else
	    {
	      set_principal_curvature(i/(double)N,j/(double)M);
	    }
	  /*	  
	  lambda1=fabs(query_principal_curvature_1());
	  lambda2=fabs(query_principal_curvature_2());

	  if(lambda1<=1.0) dR  = lambda1;
	  else dR = 1.0;
	  
	  if(lambda2<=1.0) dG  = lambda2;
	  else dG = 1.0;
	  
	  if(dR+dG>=1.0)   dB  = 0.0;
	  else dB = 1.0 - dR - dB;
	  */

	  lambda1=query_principal_curvature_1();
	  lambda2=query_principal_curvature_2();

	  if(lambda1*lambda2<=0.0) dB=1.0;
	  else dB=0.0;

	  lambda1=fabs(lambda1);
	  lambda2=fabs(lambda2);

	  if(lambda1<=1.0) dR  = lambda1;
	  else dR = 1.0;
	  
	  if(lambda2<=1.0) dG  = lambda2;
	  else dG = 1.0;

#ifdef  REFLECTION_MAP

	  if(!(((int)(30.0/M_PI*atan2(NS[0],NS[2])))%3))
	    {
	      dR=0.2;
	      dG=0.2;
	      dB=0.2;
	    }
#endif	  
	  red   = (unsigned char)(255.0 * dR);
	  green = (unsigned char)(255.0 * dG);
	  blue  = (unsigned char)(255.0 * dB);
	  
	  color=256*(256*(unsigned int)blue + (unsigned int)green) + red;
	  
	  if(i==N && j==M)
	    {
	      fprintf(f,"         0xff%.6x",color);
	    }
	  else
	    {
	      fprintf(f,"         0xff%.6x,",color);
	    }
	}
      fprintf(f,"\n");
    }

  fprintf(f,"]\n");
  fprintf(f,"    }\n");
  fprintf(f,"    MaterialBinding {\n");
  fprintf(f,"	value	PER_VERTEX_INDEXED\n");
  fprintf(f,"    }\n");
  fprintf(f,"    Coordinate3 {\n");
  fprintf(f,"	point	[\n");

  for(i=0;i<=N;i++)
    {
      for(j=0;j<=M;j++)
	{
	  if(i==N && j==M)
	    {
	      fprintf(f,"         %e %e %e",
		      evaluate_S(i/(double)N,j/(double)M, 0),
		      evaluate_S(i/(double)N,j/(double)M, 1),
		      evaluate_S(i/(double)N,j/(double)M, 2));
	    }
	  else
	    {
	      fprintf(f,"         %e %e %e,",
		      evaluate_S(i/(double)N,j/(double)M, 0),
		      evaluate_S(i/(double)N,j/(double)M, 1),
		      evaluate_S(i/(double)N,j/(double)M, 2));
	    }
	  fprintf(f,"\n");
	}
    }

  fprintf(f,"    ]}\n");
  fprintf(f,"    QuadMesh {\n");
  fprintf(f,"	verticesPerColumn	%d\n",N+1);
  fprintf(f,"	verticesPerRow	%d\n",M+1);
  fprintf(f,"    }\n");
  fprintf(f,"}\n");

  fclose(f);
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// compute Kp matrix and from that get the energy approximation

double Patch::KpEnergy(void)
{
  cerr << "Error: double Patch::KpEnergy(void) not defined" << endl;
  exit(1);
}

//////////////////////////////////////////////////////////////////////////////
/*
double Patch::grad_KpEnergy(int coord, int i)
{
  cerr << "Error: double Patch::grad_KpEnergy(int coord, int i) not defined" << endl;
  exit(1);
}

//////////////////////////////////////////////////////////////////////////////

double Patch::Hess_KpEnergy(int coord, int i)
{
  cerr << "Error: double Patch::Hess_KpEnergy(int coord, int i) not defined" << endl;
  exit(1);
}
*/
//////////////////////////////////////////////////////////////////////////////

double Patch::KmEnergy(void)
{
  cerr << "Error: double Patch::KmEnergy(void) not defined" << endl;
  exit(1);
}

//////////////////////////////////////////////////////////////////////////////
/*
double Patch::grad_KmEnergy(int coord, int i)
{
  cerr << "Error: double Patch::grad_KmEnergy(int coord, int i) not defined" << endl;
  exit(1);
}

//////////////////////////////////////////////////////////////////////////////

double Patch::Hess_KmEnergy(int coord, int i)
{
  cerr << "Error: double Patch::Hess_KmEnergy(int coord, int i) not defined" << endl;
  exit(1);
}
*/
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
