#include <stdlib.h>                      // System includes
#include <iostream.h>                    // 

#include "coord_double.h"    
#include "compcol_double.h"              // Compressed column matrix header
#include "iohb_double.h"                 // Harwell-Boeing matrix I/O header
#include "mvblasd.h"                     // MV_Vector level 1 BLAS
#include "icpre_double.h"              // Diagonal preconditioner

#include "cg.h"                          // IML++ CG template

int
main(int argc, char * argv[])
{
  if (argc < 2) {
    cerr << "Usage: " << argv[0] << " dimension " << endl;
    exit(-1);
  }

  int dim = 10;
  double tol = 1.e-6;                    // Convergence tolerance
  int result, maxit = 100000;               // Maximum iterations
  int i,j;

  dim = atoi(argv[1]);

  cout << "dim=" << dim << endl;

  double a[dim*dim];
  int    r[dim*dim];
  int    c[dim*dim];

  for(i=0;i<dim;i++)
    for(j=0;j<dim;j++) 
      {
	a[dim*i+j]=1.0/(i+j+1.0);
	r[dim*i+j]=i;
	c[dim*i+j]=j;
      }

  Coord_Mat_double A(dim,dim,dim*dim,a,r,c); // Create a matrix
  CompCol_Mat_double AA;

  AA=A;

  VECTOR_double b(dim, 1.0);             // Create rhs, solution vectors
  VECTOR_double x(dim, 0.0);             

  ICPreconditioner_double D(AA);        // Create diagonal preconditioner

  result = CG(AA, x, b, D, maxit, tol);   // Solve system

  cout << x;

  cout << "CG flag = " << result << endl;
  cout << "iterations performed: " << maxit << endl;
  cout << "tolerance achieved  : " << tol << endl;

  return result;
}
