//////////////////////////////////////////////////////////////////////////
//
// Tree.cpp, Ilja Friedel, 08/31/02
//
//////////////////////////////////////////////////////////////////////////

#include "Tree.h"

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

int TreeNode::degree(void)
{
  cerr << "Error: int TreeNode::degree(void)" << endl;
  exit(1);
  return(0);
}

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// constructor/destructor

Tree::~Tree()
{

}

//////////////////////////////////////////////////////////////////////////

Tree::Tree()
{
  root=NULL;
  N=0;
}

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// private functions

TreeNode* Tree::insert_root(Priority *P)
{
  // for right now
  assert(root==NULL);
  assert(N==0);

  root=new TreeNode(P);
  assert(root);
  N++;
  return(root);
}

//////////////////////////////////////////////////////////////////////////

TreeNode* Tree::insert_child(TreeNode *TN, Priority *P)
{
  if(TN->lchild)
    {
      return(insert_rsibling(TN->lchild,P));
    }
  else
    {
      TN->lchild         = new TreeNode(P);
      TN->lchild->parent = TN;
      N++;
      return(TN->lchild);
    }
}

//////////////////////////////////////////////////////////////////////////

TreeNode* Tree::insert_rsibling(TreeNode *TN, Priority *P)
{
  TN->next=new TreeNode(TN, TN->next, P);
  TN->next->parent=TN->parent;
  N++;
  return(TN->next);
}

//////////////////////////////////////////////////////////////////////////

void Tree::remove(TreeNode *TN)
{
  // for now TN must have no children (can change later)
  assert(TN->lchild==NULL);

  // TN is on the topmost level
  if(TN->parent==NULL)
    {
      // again, no kids allowed
      assert(TN->lchild==NULL);

      // join siblings
      TN->prev->next=TN->next;
      TN->next->prev=TN->prev;
      
      // if TN is root, take sibling as root
      if(root==TN)
	{
	  root=TN->next;
	}
      
      // if TN is still root, must be last Node in Tree
      if(root==TN)
	{
	  assert(N==1);
	  root=NULL;
	}

      delete TN;
      N--;
      return;
    }

  assert(TN->parent!=NULL);

  // if TN is lchild of parent make sibling of TN new lchild
  if(TN==TN->parent->lchild)
    {
      TN->parent->lchild=TN->next;
    }
  // if TN is the only child of parent, then link the kids of TN as kids of parent
  if(TN->next==TN)
    {
      TN->parent->lchild=TN->lchild;
    }
     
  // if TN has no children, join the siblings of TN
  if(TN->lchild==NULL)
    {
      TN->prev->next=TN->next;
      TN->next->prev=TN->prev;
    }
  else // link kids of TN as siblings
    {
      // again, no kids (doesn't work yet)
      assert(TN->lchild==NULL);

      TN->prev->next         = TN->lchild;
      TN->next->prev         = TN->lchild->prev;
      TN->lchild->prev       = TN->prev;
      TN->lchild->prev->next = TN->next;

      // ************** need to set parent of all the new kids! *************
    }

  // get rid of TN
  delete TN;
  N--;
}

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// public functions

int Tree::size(void)
{
  assert(N>=0);
  return(N);
}

//////////////////////////////////////////////////////////////////////////

int Tree::degree(TreeNode *TN)
{
  TreeNode *TN1, *TN2;
  int d;

  assert(TN);
  if(TN->lchild==NULL) return(0);
  TN1=TN->lchild;
  TN2=TN->lchild;
  
  d=0;
  do
    {
      d++;
      TN1=TN1->next;
    }
  while(TN1!=TN2);

  return(d);
}

//////////////////////////////////////////////////////////////////////////

TreeNode* Tree::FirstElement(void)
{
  return(root);
}

//////////////////////////////////////////////////////////////////////////

TreeNode* Tree::NextElement(TreeNode *TN)
{
  cerr << "Error: TreeNode* TreeNode::NextElement(TreeNode *TN)" << endl;
  exit(1);
  return(TN);
}

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
